from __future__ import division
from past.utils import old_div
import math

import astropy.units as astropy_units
import numpy as np
from scipy.special import  erfcinv, erf

from astromodels.functions.function import Function1D, FunctionMeta, ModelAssertionViolation
from future.utils import with_metaclass


deg2rad = old_div(np.pi,180.)
rad2deg = old_div(180.,np.pi)

# noinspection PyPep8Naming
class Gaussian(with_metaclass(FunctionMeta, Function1D)):
    r"""
    description :

        A Gaussian function

    latex : $ K \frac{1}{\sigma \sqrt{2 \pi}}\exp{\frac{(x-\mu)^2}{2~(\sigma)^2}} $

    parameters :

        F :

            desc : Integral between -inf and +inf. Fix this to 1 to obtain a Normal distribution
            initial value : 1

        mu :

            desc : Central value
            initial value : 0.0

        sigma :

            desc : standard deviation
            initial value : 1.0
            min : 1e-12

    tests :
        - { x : 0.0, function value: 0.3989422804014327, tolerance: 1e-10}
        - { x : -1.0, function value: 0.24197072451914337, tolerance: 1e-9}

    """

    # Place this here to avoid recomputing it all the time

    __norm_const = old_div(1.0, (math.sqrt(2 * np.pi)))

    def _setup(self):

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):

        # The normalization is the integral from -inf to +inf, i.e., has dimensions of
        # y_unit * x_unit
        self.F.unit = y_unit * x_unit

        # The mu has the same dimensions as the x
        self.mu.unit = x_unit

        # sigma has the same dimensions as x
        self.sigma.unit = x_unit

    # noinspection PyPep8Naming
    def evaluate(self, x, F, mu, sigma):

        norm = old_div(self.__norm_const, sigma)

        return F * norm * np.exp(old_div(-np.power(x - mu, 2.), (2 * np.power(sigma, 2.))))

    def from_unit_cube(self, x):
        """
        Used by multinest

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        mu = self.mu.value
        sigma = self.sigma.value

        sqrt_two = 1.414213562

        if x < 1e-16 or (1 - x) < 1e-16:

            res = -1e32

        else:

            res = mu + sigma * sqrt_two * erfcinv(2 * (1 - x))

        return res

class Truncated_gaussian(with_metaclass(FunctionMeta, Function1D)):
    r"""
    description :

        A  truncated Gaussian function defined on the interval between the lower_bound (a) and upper_bound (b)

    latex : $\begin{split}f(x;\mu,\sigma,a,b)=\frac{\frac{1}{\sigma} \phi\left( \frac{x-\mu}{\sigma} \right)}{\Phi\left( \frac{b-\mu}{\sigma} \right) - \Phi\left( \frac{a-\mu}{\sigma} \right)}\\\phi\left(z\right)=\frac{1}{\sqrt{2 \pi}}\exp\left(-\frac{1}{2}z^2\right)\\\Phi\left(z\right)=\frac{1}{2}\left(1+erf\left(\frac{z}{\sqrt(2)}\right)\right)\end{split}$

    parameters :

        F :

            desc : Integral between -inf and +inf. Fix this to 1 to obtain a Normal distribution
            initial value : 1

        mu :

            desc : Central value
            initial value : 0.0

        sigma :

            desc : standard deviation
            initial value : 1.0
            min : 1e-12

        lower_bound :

            desc: lower bound of gaussian, setting to -np.inf results in half normal distribution
            initial value : -1.

        upper_bound :

            desc: upper bound of gaussian  setting to np.inf results in half normal distribution
            initial value : 1.


    tests :
        - { x : 0.0, function value: 0.3989422804014327, tolerance: 1e-10}
        - { x : -1.0, function value: 0.24197072451914337, tolerance: 1e-9}

    """

    # Place this here to avoid recomputing it all the time

    __norm_const = old_div(1.0, (math.sqrt(2 * np.pi)))

    def _setup(self):

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):

        # The normalization is the integral from -inf to +inf, i.e., has dimensions of
        # y_unit * x_unit
        self.F.unit = y_unit * x_unit

        # The mu has the same dimensions as the x
        self.mu.unit = x_unit

        # The lower_bound has the same dimensions as the x
        self.lower_bound.unit = x_unit

        # The upper_bound has the same dimensions as the x
        self.upper_bound.unit = x_unit

        # sigma has the same dimensions as x
        self.sigma.unit = x_unit


    # noinspection PyPep8Naming
    def evaluate(self, x, F, mu, sigma, lower_bound, upper_bound):


        # phi is in unitless, so we need to do this trick
        # to keep the units right

        norm = old_div(self.__norm_const, sigma)

        phi = np.zeros(x.shape) * F * norm * 0.
        idx = (x >= lower_bound) & (x <= upper_bound)

        sqrt_two = 1.414213562

        # precalculate the arguments to the CDF

        lower_arg = old_div((lower_bound - mu), sigma)
        upper_arg = old_div((upper_bound - mu), sigma)



        # the typical gaussian functions

        phi[idx] = np.exp(old_div(-np.power(x[idx] - mu, 2.), (2 * np.power(sigma, 2.)))) * F * norm

        # the denominator is a function of the CDF


        if isinstance(F, astropy_units.Quantity):
            # erf cannot accept units

            upper_arg = upper_arg.value
            lower_arg = lower_arg.value

        theta_lower = 0.5 + 0.5 * erf(old_div(lower_arg, sqrt_two))

        theta_upper = 0.5 + 0.5 * erf(old_div(upper_arg, sqrt_two))




        return old_div(phi, (theta_upper - theta_lower))

    def from_unit_cube(self, x):

        mu = self.mu.value
        sigma = self.sigma.value
        lower_bound = self.lower_bound.value
        upper_bound = self.upper_bound.value

        sqrt_two = 1.414213562

        if x < 1e-16 or (1 - x) < 1e-16:
            res = -1e32

        # precalculate the arguments to the  CDF

        lower_arg = old_div((lower_bound - mu), sigma)
        upper_arg = old_div((upper_bound - mu), sigma)

        theta_lower = 0.5 + 0.5 * erf(old_div(lower_arg, sqrt_two))

        theta_upper = 0.5 + 0.5 * erf(old_div(upper_arg, sqrt_two))

        # now precalculate the argument to the Inv. CDF

        arg = theta_lower + x * (theta_upper - theta_lower)

        out = mu + sigma * sqrt_two * erfcinv(2 * (1 - arg))
        
        return np.clip(out, lower_bound, upper_bound)

class Cauchy(with_metaclass(FunctionMeta, Function1D)):
    r"""
    description :

        The Cauchy distribution

    latex : $ K \frac{1}{ \gamma \pi} \left[ \frac{\gamma^2}{(x-x_0)^2 + \gamma^2}  \right] $

    parameters :

        K :

            desc : Integral between -inf and +inf. Fix this to 1 to obtain a Cauchy distribution
            initial value : 1

        x0 :

            desc : Central value
            initial value : 0.0

        gamma :

            desc : standard deviation
            initial value : 1.0
            min : 1e-12

    tests :
        - { x : 0.0, function value: 0.3989422804014327, tolerance: 1e-10}
        - { x : -1.0, function value: 0.24197072451914337, tolerance: 1e-9}

    """

    # Place this here to avoid recomputing it all the time

    __norm_const = old_div(1.0, (math.sqrt(2 * np.pi)))

    def _setup(self):
        self._is_prior = True

    def _set_units(self, x_unit, y_unit):
        # The normalization is the integral from -inf to +inf, i.e., has dimensions of
        # y_unit * x_unit
        self.K.unit = y_unit * x_unit

        # The mu has the same dimensions as the x
        self.x0.unit = x_unit

        # sigma has the same dimensions as x
        self.gamma.unit = x_unit

    # noinspection PyPep8Naming
    def evaluate(self, x, K, x0, gamma):
        norm = old_div(1, (gamma * np.pi))

        gamma2 = gamma * gamma

        return K * norm * gamma2 / ((x - x0) * (x - x0) + gamma2)

    def from_unit_cube(self, x):
        """
        Used by multinest

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        x0 = self.x0.value
        gamma = self.gamma.value

        half_pi = 1.57079632679

        res = np.tan(np.pi * x - half_pi) * gamma + x0

        return res


class Cosine_Prior(with_metaclass(FunctionMeta, Function1D)):
    r"""
    description :

        A function which is constant on the interval angular interval of cosine

    latex : $\cos(x)$

    parameters :

        lower_bound :

            desc : Lower bound for the interval
            initial value : -90
            min : -np.inf
            max : np.inf

        upper_bound :

            desc : Upper bound for the interval
            initial value : 90
            min : -np.inf
            max : np.inf


        value :

            desc : Value in the interval
            initial value : 1.0



    """


    def _setup(self):

        self._fixed_units = (astropy_units.dimensionless_unscaled,astropy_units.dimensionless_unscaled)

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):


        # this prior needs to use the fixed units and
        # they do not need to be converted as they have no
        # dimension

        x_unit = self._fixed_units[0]
        y_unit = self._fixed_units[1]

        # Lower and upper bound has the same unit as x

        self.lower_bound.unit = x_unit
        self.upper_bound.unit = x_unit

        # value has the same unit as y
        self.value.unit = y_unit

    def has_fixed_units(self):

        return True

    def evaluate(self, x, lower_bound, upper_bound,value):
        # The value * 0 is to keep the units right

        result = np.zeros(x.shape) * value * 0

        idx = (x >= lower_bound) & (x <= upper_bound)

        norm = (np.sin(deg2rad*(upper_bound)) - np.sin(deg2rad*(lower_bound))) * 57.29577795


        result[idx] = value * np.cos(deg2rad*( x[idx] )) / norm

        return result


    def from_unit_cube(self, x):
        """
        Used by multinest

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """
        cosdec_min = np.cos(deg2rad*(90.0 + self.lower_bound.value))
        cosdec_max = np.cos(deg2rad*(90.0 + self.upper_bound.value))

        v = x * (cosdec_max - cosdec_min)
        v += cosdec_min

        v = np.clip(v, -1.0, 1.0)
        # Now this generates on [0,pi)
        dec = np.arccos(v)

        # convert to degrees
        dec = rad2deg * dec
        # now in range [-90,90.0)
        dec -= 90.0

        return dec


class Log_normal(with_metaclass(FunctionMeta, Function1D)):
    r"""
       description :

           A log normal function

       latex : $ K \frac{1}{ x \sigma \sqrt{2 \pi}}\exp{\frac{(\log x/piv - \mu/piv)^2}{2~(\sigma)^2}} $

       parameters :

           F :
               desc : Integral between 0and +inf. Fix this to 1 to obtain a log Normal distribution
               initial value : 1

           mu :

               desc : Central value
               initial value : 0.0

           sigma :

               desc : standard deviation
               initial value : 1.0
               min : 1e-12

           piv :
               desc : pivot. Leave this to 1 for a proper log normal distribution
               initial value : 1.0
               fix : yes

       tests :
           - { x : 0.0, function value: 0.3989422804014327, tolerance: 1e-10}
           - { x : -1.0, function value: 0.24197072451914337, tolerance: 1e-9}

       """

    # Place this here to avoid recomputing it all the time

    __norm_const = old_div(1.0, (math.sqrt(2 * np.pi)))

    def _setup(self):

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):

        # The normalization is the integral from -inf to +inf, i.e., has dimensions of
        # y_unit * x_unit
        self.F.unit = y_unit

        # The mu has the same dimensions as the x
        self.mu.unit = x_unit

        # The pivot has the same units as x
        self.piv.unit = x_unit

        # sigma has the same dimensions as x
        self.sigma.unit = x_unit

    # noinspection PyPep8Naming
    def evaluate(self, x, F, mu, sigma, piv):

        # The value * 0 is to keep the units right

        result = np.zeros(x.shape) * F * 0

        # The log normal is not defined if x < 0. The "0 * x" part is to conserve the units if
        # x has them, because 0 * x will be a Quantity with the same units as x
        idx = (x > 0 * x)

        result[idx] = F * self.__norm_const / (sigma / piv * x / piv) * np.exp(
            old_div(-np.power(np.log(old_div(x, piv)) - old_div(mu, piv), 2.), (2 * np.power(old_div(sigma, piv), 2.))))

        return result

    def from_unit_cube(self, x):
        """
        Used by multinest

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        mu = self.mu.value
        sigma = self.sigma.value

        sqrt_two = 1.414213562

        if x < 1e-16 or (1 - x) < 1e-16:

            res = -1e32

        else:

            res = mu + sigma * sqrt_two * erfcinv(2 * (1 - x))

        return np.exp(res)


class Uniform_prior(with_metaclass(FunctionMeta, Function1D)):
    r"""
    description :

        A function which is constant on the interval lower_bound - upper_bound and 0 outside the interval. The
        extremes of the interval are counted as part of the interval.

    latex : $ f(x)=\begin{cases}0 & x < \text{lower_bound} \\\text{value} & \text{lower_bound} \le x \le \text{upper_bound} \\ 0 & x > \text{upper_bound} \end{cases}$

    parameters :

        lower_bound :

            desc : Lower bound for the interval
            initial value : 0
            min : -np.inf
            max : np.inf

        upper_bound :

            desc : Upper bound for the interval
            initial value : 1
            min : -np.inf
            max : np.inf

        value :

            desc : Value in the interval
            initial value : 1.0

    tests :
        - { x : 0.5, function value: 1.0, tolerance: 1e-20}
        - { x : -0.5, function value: 0, tolerance: 1e-20}

    """

    def _setup(self):
        self._is_prior = True

    def _set_units(self, x_unit, y_unit):
        # Lower and upper bound has the same unit as x
        self.lower_bound.unit = x_unit
        self.upper_bound.unit = x_unit

        # value has the same unit as y
        self.value.unit = y_unit

    def evaluate(self, x, lower_bound, upper_bound, value):
        # The value * 0 is to keep the units right

        result = np.zeros(x.shape) * value * 0

        idx = (x >= lower_bound) & (x <= upper_bound)
        result[idx] = value

        return result


    def from_unit_cube(self, x):
        """
        Used by multinest

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        lower_bound = self.lower_bound.value
        upper_bound = self.upper_bound.value

        low = lower_bound
        spread = float(upper_bound - lower_bound)

        par = x * spread + low

        return par

class Log_uniform_prior(with_metaclass(FunctionMeta, Function1D)):
    r"""
    description :

        A function which is K/x on the interval lower_bound - upper_bound and 0 outside the interval. The
        extremes of the interval are NOT counted as part of the interval. Lower_bound must be >= 0.

    latex : $ f(x)=K~\begin{cases}0 & x \le \text{lower_bound} \\\frac{1}{x} & \text{lower_bound} < x < \text{upper_bound} \\ 0 & x \ge \text{upper_bound} \end{cases}$

    parameters :

        lower_bound :

            desc : Lower bound for the interval
            initial value : 1e-20
            min : 1e-30
            max : np.inf

        upper_bound :

            desc : Upper bound for the interval
            initial value : 100
            min : 1e-30
            max : np.inf

        K :

            desc : Normalization
            initial value : 1
            fix : yes

    """

    def _setup(self):

        self._is_prior = True
        self._handle_units = False

    def _set_units(self, x_unit, y_unit):
        # Lower and upper bound has the same unit as x
        self.lower_bound.unit = x_unit
        self.upper_bound.unit = x_unit
        self.K.unit = y_unit * x_unit

    def evaluate(self, x, lower_bound, upper_bound, K):
        # This makes the prior proper because it is the integral between lower_bound and upper_bound

        res = np.where((x > lower_bound) & (x < upper_bound), old_div(K, x), 0)

        if isinstance(x, astropy_units.Quantity):

            return res * self.y_unit

        else:

            return res

    def from_unit_cube(self, x):
        """
        Used by multinest

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        low = math.log10(self.lower_bound.value)
        up = math.log10(self.upper_bound.value)

        spread = up - low
        par = 10 ** (x * spread + low)

        return par




import numpy.testing as npt
import numpy as np
import h5py


from astromodels.functions.priors import *
from astromodels.functions.function import _known_functions
from astromodels.utils.data_files import _get_data_file_path

_multiplicative_models = ["PhAbs", "TbAbs", "WAbs", "APEC", "VAPEC", "EBLattenuation" ]

def test_function_values_have_not_changed():

    with h5py.File(_get_data_file_path("past_1D_values.h5"), "r") as f:

        eval_x = f["eval_values"][()]

    
    for key in _known_functions:

        this_function = _known_functions[key]

        # Test only the power law of XSpec, which is the only one we know we can test at 1 keV

        if key.find("XS")==0 or (key in _multiplicative_models):

            # An XSpec model OR EBLattenuation function. Test it only if it's a power law (the others might need other parameters during
            # initialization)

            continue

        if key.find("TemplateModel")==0:

            # The TemplateModel function has its own test

            continue

    #        if key.find("Synchrotron")==0:

            # Naima Synchtron function should have its own test

    #            continue

        if this_function._n_dim == 1:

            print("testing %s ..." % key)


            
            
            func = this_function()
            
            new_values = np.atleast_1d(func(eval_x))

            with h5py.File(_get_data_file_path("past_1D_values.h5"), "r") as f:
                if key not in f.keys():

                    raise RuntimeError("the function %s does not exist in the past data. You must run a script to add it" %key)
                
                old_values = f[key][()]

            
            
            npt.assert_almost_equal(new_values, old_values)
    



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import object
import astropy.units as u
import pytest
from astromodels.functions.priors import Uniform_prior, Log_uniform_prior
from astromodels.functions.functions import  Powerlaw

__author__ = 'giacomov'

from astromodels.core.parameter import Parameter, SettingOutOfBounds, \
    CannotConvertValueToNewUnits, NotCallableOrErrorInCall, IndependentVariable, ParameterMustHaveBounds
from astromodels.functions.functions import Line
from astromodels.core.parameter_transformation import LogarithmicTransformation

def test_default_constructor():

    p = Parameter('test_parameter', 1.0, desc='Description')

    assert p.min_value is None
    assert p.max_value is None
    assert p.value == 1.0
    assert isinstance(p.delta, float)
    assert p.name == 'test_parameter'
    assert p.description == 'Description'
    assert p.fix == False
    assert p.free == True
    assert p.has_prior() == False
    assert p.prior is None

    assert p.unit == u.dimensionless_unscaled

    # Test that we cannot call a parameter with a name with spaces in it
    with pytest.raises(AssertionError):

        _ = Parameter('test parameter 2', 1.0)

    # Test some failures cases
    with pytest.raises(ValueError):

        _ = Parameter('test', 'pippo')

    with pytest.raises(ValueError):
        _ = Parameter('test', 1.0, min_value='a')

    with pytest.raises(ValueError):

        _ = Parameter('test', 1.0, max_value='b')

    with pytest.raises(TypeError):

        _ = Parameter('test', 1.0, delta='b')

    p.display()


def test_default_constructor_units():

    p = Parameter('test_parameter', 1.0 * u.keV, desc='Description')

    assert p.min_value is None
    assert p.max_value is None
    assert p.value == 1.0
    assert isinstance(p.delta, float)
    assert p.name == 'test_parameter'
    assert p.description == 'Description'
    assert p.fix == False
    assert p.free == True
    assert p.has_prior() == False
    assert p.prior is None

    assert p.unit == u.keV

    p.display()


def test_constructor_complete():

    p = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0, delta=0.2, desc='test',
                  free=False, unit=u.MeV, prior=Uniform_prior(), is_normalization=True,)

    assert p.min_value == -5.0
    assert p.max_value == 5.0
    assert p.value == 1.0
    assert p.delta == 0.2
    assert p.name == 'test_parameter'
    assert p.description == 'test'
    assert p.fix == True
    assert p.free == False
    assert p.has_prior() == True

    assert p.unit == u.MeV

    assert p.is_normalization

    p.display()

def test_constructor_with_transform():

    p = Parameter('test_parameter', 1.0, min_value=0.1, max_value=5.0, delta=0.2, desc='test',
                  free=False, unit=u.MeV, prior=Uniform_prior(), is_normalization=True,)

    assert p.min_value == 0.1
    assert p.max_value == 5.0
    assert p.value == 1.0
    assert p.delta == 0.2
    assert p.name == 'test_parameter'
    assert p.description == 'test'
    assert p.fix == True
    assert p.free == False
    assert p.has_prior() == True

    assert p.unit == u.MeV

    assert p.is_normalization

    assert p.has_transformation
    
    p.display()


    
def test_conflicting_units_in_initial_value_and_unit_keyword():

    p = Parameter('test_parameter', 1.0 * u.keV, desc='Description', unit=u.MeV)

    assert p.min_value is None
    assert p.max_value is None
    assert p.value == 1.0e-3
    assert isinstance(p.delta, float)
    assert p.name == 'test_parameter'
    assert p.description == 'Description'
    assert p.fix == False
    assert p.free == True
    assert p.has_prior() == False
    assert p.prior is None

    assert p.unit == u.MeV

    p.display()


def test_constructor_with_boundaries():

    p = Parameter('test_parameter', 1.0, min_value=-5, max_value=5)

    assert p.min_value == -5
    assert p.max_value == 5

    p.display()


def test_constructor_with_delta():

    p = Parameter('test_parameter', 1.0, delta=0.3)

    assert p.delta == 0.3

    p.display()


def test_constructor_with_units():

    p = Parameter('test_parameter', 1.0, unit=u.keV)

    assert p.unit == u.keV

    p.display()


def test_set_no_units():

    p = Parameter('test_parameter',1.0)

    p.value = 25.4

    assert p.value == 25.4

    p.display()


def test_set_within_bounds_no_units():

    p = Parameter('test_parameter',1.0, min_value = -2.0, max_value = 2.0)

    p.value = 1.5

    assert p.value == 1.5

    p.display()


def test_set_outside_bounds_no_units():

    p = Parameter('test_parameter',1.0, min_value = -2.0, max_value = 2.0)

    with pytest.raises(SettingOutOfBounds):

        p.value = -10.0

    with pytest.raises(SettingOutOfBounds):

        p.value = 10.0

    p.display()


def test_set_units():

    p = Parameter('test_parameter',1.0, unit=u.keV)

    p.value = 3.0 * u.MeV

    assert p.value == 3000.0

    with pytest.raises(u.UnitConversionError):

        p.value = 3.0 * u.cm

    with pytest.raises(CannotConvertValueToNewUnits):

        p.unit = u.cm

    with pytest.raises(CannotConvertValueToNewUnits):

        p.unit = u.dimensionless_unscaled

    p.unit = u.MeV

    assert p.unit == u.MeV

    p.display()


def test_set_within_bounds_units():

    p = Parameter('test_parameter',1.0 * u.keV, min_value = -2.0 * u.MeV, max_value = 2.0 * u.MeV, unit=u.keV)

    p.value = 1.2 * u.MeV

    assert p.value == 1200.0

    p.display()


def test_set_outside_bounds_units():

    p = Parameter('test_parameter', 1.0 * u.keV, min_value = -2.0 * u.MeV, max_value = 2.0 * u.MeV, unit=u.keV)

    with pytest.raises(SettingOutOfBounds):

        p.value = -10.0 * u.MeV

    with pytest.raises(SettingOutOfBounds):

        p.value = 10.0 * u.MeV

    p.display()


def test_set_bounds_nounits():

    p = Parameter('test_parameter', 1.0)

    p.bounds = (-2.0 ,2.0)

    assert p.min_value == -2.0
    assert p.max_value == 2.0

    p.display()

    with pytest.warns(RuntimeWarning):

        p.value = 1.0
        p.min_value = 2.0


def test_set_bounds_units():

    p = Parameter('test_parameter', 1.0 * u.keV)

    p.bounds = (-2.0 * u.MeV, 2.0 * u.eV)

    assert p.min_value == -2000
    assert p.max_value == 2e-3

    p.display()


def test_set_delta_nounits():

    p = Parameter('test_parameter', 1.0)

    p.delta = 0.5

    assert p.delta == 0.5

    p.display()


def test_set_delta_units():

    p = Parameter('test_parameter', 1.0, unit='GeV')

    p.delta = 500 * u.MeV

    assert p.delta == 0.5

    p.display()


def test_duplicate():

    p1 = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0, delta=0.2, desc='test', free=False, unit='MeV')

    p2 = p1.duplicate()

    assert p1.to_dict() == p2.to_dict()

    p1.display()
    p2.display()


def test_get_randomized_value():

    # Test randomization no boundaries (normal distribution)
    p1 = Parameter('test_parameter', 1.0)

    val2 = p1.get_randomized_value(0.1)

    assert isinstance(val2, float)

    # Test the randomized value with truncated normal, i.e., with boundaries

    p2 = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0, delta=0.2, desc='test', free=False, unit='MeV')

    val1 = p2.get_randomized_value(0.1)

    assert p2.min_value <= val1 <= p2.max_value

    # Test the same but with a large variance

    val1 = p2.get_randomized_value(10.0)

    assert p2.min_value <= val1 <= p2.max_value

    p2.min_value = None

    val1 = p2.get_randomized_value(10.0)
    assert val1 <= p2.max_value

    p2.min_value = -5.0
    p2.max_value = None

    val1 = p2.get_randomized_value(10.0)
    assert val1 >= p2.min_value


def test_set_remove_minimum():
    p1 = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0, delta=0.2, desc='test', free=False, unit='MeV')

    p1.remove_minimum()

    assert p1.min_value == None

    p1.value = -1000.0

    assert p1.value == -1000.0


def test_set_remove_maximum():

    p1 = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0, delta=0.2, desc='test', free=False, unit='MeV')

    p1.remove_maximum()

    assert p1.max_value == None

    p1.value = 1000.0

    assert p1.value == 1000.0


def test_set_auxiliary_variable():

    p1 = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0, delta=0.2, desc='test', free=False, unit='MeV')

    x = Parameter('aux_variable', 1.0)

    # ax + b

    law = Line()
    law.a = 1.0
    law.b = 2.0

    p1.add_auxiliary_variable(x, law)

    assert p1.has_auxiliary_variable() == True

    assert p1.value == 3.0

    assert p1.free == False

    x.value = 4.0

    assert p1.value == 6.0

    # Check that assigning to the parameter doesn't produce any effect
    p1.value = -1.0

    assert p1.value == 6.0

    # Now check errors reporting
    with pytest.raises(AttributeError):

        p1.add_auxiliary_variable(1.0, law)

    # Now add it twice to verify that it overwrites it
    p1.add_auxiliary_variable(x, law)
    p1.add_auxiliary_variable(x, law)

    p1.display()



def test_remove_auxiliary_variable():

    p1 = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0, delta=0.2, desc='test', free=False, unit='MeV')

    x = Parameter('aux_variable', 1.0)

    # ax + b

    law = Line()
    law.a = 1.0
    law.b = 2.0

    p1.add_auxiliary_variable(x, law)

    assert p1.value == 3.0

    x.value = 4.0

    assert p1.value == 6.0

    p1.remove_auxiliary_variable()

    assert p1.has_auxiliary_variable() == False

    p1.value = -1.0

    assert p1.value == -1.0

    with pytest.warns(RuntimeWarning):

        p1.remove_auxiliary_variable()


def test_callback():

    p1 = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0, delta=0.2, desc='test', free=False, unit='MeV')

    class Callback(object):

        def __init__(self):

            self._control_value = None

        def __call__(self, p):

            assert p == p1

            self._control_value = p1.value

    working_callback = Callback()

    p1.add_callback(working_callback)

    # Test the callback
    p1.value = 2.0
    assert working_callback._control_value == p1.value

    def not_working_callback():

        # Wrong calling sequence
        pass

    p1.add_callback(not_working_callback)

    # This should work because we do not change the parameter, so the callback
    # does not get called
    p1.value = 2.0

    # This should instead raise because we do change the value
    with pytest.raises(NotCallableOrErrorInCall):

        p1.value = 3.0

    p1.empty_callbacks()

    assert len(p1._callbacks) == 0


def test_to_dict():

    p1 = IndependentVariable('time', 1.0, min_value=-5.0, max_value=5.0, desc='test', unit='MeV')

    repr = p1.to_dict(minimal=False)

    assert len(list(repr.keys())) == 5

    repr2 = p1.to_dict(minimal=True)

    assert len(list(repr2.keys())) == 1
    assert 'value' in repr2

    assert repr2['value'] == p1.value

    p = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0,
                   delta=0.2, desc='test', free=False, unit='MeV')

    p.to_dict()
    p.to_dict(minimal=True)

    p = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0,
                  delta=0.2, desc='test', free=False, unit='MeV', prior=Log_uniform_prior())

    p.to_dict()
    p.to_dict(minimal=True)


def test_independent_variable_representation():

    p1 = IndependentVariable('time', 1.0, min_value=-5.0, max_value=5.0, desc='test', unit='MeV')

    print(p1._repr__base(False))
    print(p1._repr__base(True))


def test_prior():

    p1 = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0,
                   delta=0.2, desc='test', free=False, unit='MeV')

    my_prior = Uniform_prior()

    p1.prior = my_prior

    assert my_prior == p1.prior

    custom_prior = lambda x:x**2

    with pytest.raises(NotCallableOrErrorInCall):

        p1.prior = custom_prior

    invalid_prior = lambda x,y: x*y

    with pytest.raises(NotCallableOrErrorInCall):

        p1.prior = invalid_prior

    # Test the set_uninformative_prior method
    p1.min_value = None
    p1.max_value = 100.0

    with pytest.raises(ParameterMustHaveBounds):

        p1.set_uninformative_prior(Uniform_prior)

    p1.min_value = 0.0
    p1.max_value = None

    with pytest.raises(ParameterMustHaveBounds):

        p1.set_uninformative_prior(Uniform_prior)

    p1.min_value = 0.0
    p1.max_value = 100.0

    p1.set_uninformative_prior(Uniform_prior)

    # Log-uniform cannot be used if minimum is 0.0
    with pytest.raises(SettingOutOfBounds):

        p1.set_uninformative_prior(Log_uniform_prior)

    p1.min_value = 1.0
    p1.set_uninformative_prior(Log_uniform_prior)


def test_remove_prior():

    p1 = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0,
                   delta=0.2, desc='test', free=False, unit='MeV')

    my_prior = Uniform_prior()

    p1.prior = my_prior

    assert p1.has_prior()==True

    # Now remove it
    p1.prior = None

    assert p1.has_prior()==False


def test_as_quantity():

    p = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0,
                  delta=0.2, desc='test', free=False, unit='MeV')

    assert isinstance(p.as_quantity, u.Quantity)
    assert p.as_quantity.to("keV").value == 1000.0


def test_in_unit_of():

    p = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0,
                  delta=0.2, desc='test', free=False, unit='MeV')

    assert p.in_unit_of(u.keV) == 1000.0
    assert p.in_unit_of(u.keV, as_quantity=True).to("MeV").value == 1.0


class Callback(object):

    def __init__(self):

        self._control_value = None

    def __call__(self, p):

        self._control_value = p.value


def test_pickle():

    from astromodels.core.cpickle_compatibility_layer import cPickle

    p_orig = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0, delta=0.2, desc='test',
                       free=False, unit=u.MeV, prior=Uniform_prior())

    # Add a callback

    working_callback = Callback()

    p_orig.add_callback(working_callback)

    # Now pickle and unpickle

    d = cPickle.dumps(p_orig)

    p = cPickle.loads(d)

    # Check that everything is fine

    assert p.min_value == -5.0
    assert p.max_value == 5.0
    assert p.value == 1.0
    assert p.delta == 0.2
    assert p.name == 'test_parameter'
    assert p.description == 'test'
    assert p.fix == True
    assert p.free == False
    assert p.has_prior() == True

    assert p.unit == u.MeV

    # Test the callback
    p.value = 2.0

    callback = p.get_callbacks()[0]

    assert callback._control_value == p.value

def test_links_and_pickle():

    import pickle

    p_orig = Parameter('test_parameter', 1.0, min_value=-5.0, max_value=5.0, delta=0.2, desc='test',
                       free=False, unit=u.MeV, prior=Uniform_prior())

    # Test the linkinking and pickle

    # Add a link
    x = Parameter('aux_variable', 1.0)

    # ax + b

    law = Line()
    law.a = 1.0
    law.b = 2.0

    p_orig.add_auxiliary_variable(x, law)

    # Now pickle and unpickle

    d = pickle.dumps(p_orig)

    p = pickle.loads(d)

    assert p.has_auxiliary_variable() == True

    assert p.value == 3.0

    assert p.free == False

    p.auxiliary_variable[0].value = 4.0

    assert p.value == 6.0

    # Check that assigning to the parameter doesn't produce any effect
    p.value = -1.0

    assert p.value == 6.0

def test_internal_setting():


    p = Parameter('test_parameter',1.0)

    p._set_internal_value(5.)


    p._get_internal_value()

    p.display()

    p._get_internal_min_value()
    p._get_internal_max_value()

def test_internal_delta():

    p = Parameter('test_parameter',1.0)
    p.min_value = None
    p.max_value = None

    p._get_internal_delta()


    p = Parameter('test_parameter', 1.0, min_value=0.1, max_value=5.0, delta=0.2, desc='test',
                  free=False, unit=u.MeV, prior=Uniform_prior(), is_normalization=True, transformation=LogarithmicTransformation())

    p._get_internal_delta()

    with pytest.raises(AssertionError):
        p = Parameter('test_parameter', 1.0, min_value=-1., max_value=5.0, delta=0.2, desc='test',
                  free=False, unit=u.MeV, prior=Uniform_prior(), is_normalization=True, transformation=LogarithmicTransformation())


    p = Parameter('test_parameter', 1.0, min_value=None, max_value=5.0, delta=0.2, desc='test',
                  free=False, unit=u.MeV, prior=Uniform_prior(), is_normalization=True, transformation=LogarithmicTransformation())



    p._get_internal_delta()



__author__ = 'giacomov'

import collections

from astropy import coordinates

from astromodels.core.parameter import Parameter
from astromodels.core.tree import Node


class WrongCoordinatePair(ValueError):
    pass


class IllegalCoordinateValue(ValueError):
    pass


class WrongCoordinateSystem(ValueError):
    pass


class SkyDirection(Node):
    """
    This is essentially a wrapper around astropy.coordinates.SkyCoord with a possibility for
    being serialized and deserialized with YAML.
    """

    def __init__(self, ra=None, dec=None, l=None, b=None, equinox='J2000'):
        """

        :param ra: Right Ascension in degrees
        :param dec: Declination in degrees
        :param l: Galactic latitude in degrees
        :param b: Galactic longitude in degrees
        :param equinox: string
        :return:
        """

        self._equinox = equinox

        # Create the node

        Node.__init__(self, 'position')

        # Check that we have the right pairs of coordinates

        if ra is not None and dec is not None:

            # This goes against duck typing, but it is needed to provide a means of initiating this class
            # with either Parameter instances or just floats

            # Try to transform it to float, if it works than we transform it to a parameter

            ra = self._get_parameter_from_input(ra, 0, 360, 'ra','Right Ascension')

            dec = self._get_parameter_from_input(dec, -90, 90, 'dec','Declination')

            self._coord_type = 'equatorial'

            self._add_child(ra)
            self._add_child(dec)

        elif l is not None and b is not None:

            # This goes against duck typing, but it is needed to provide a means of initiating this class
            # with either Parameter instances or just floats

            # Try to transform it to float, if it works than we transform it to a parameter

            l = self._get_parameter_from_input(l, 0, 360, 'l','Galactic longitude')

            b = self._get_parameter_from_input(b, -90, 90, 'b','Galactic latitude')

            self._coord_type = 'galactic'
            self._add_child(l)
            self._add_child(b)

        else:

            raise WrongCoordinatePair("You have to specify either (ra, dec) or (l, b).")

    @staticmethod
    def _get_parameter_from_input(number_or_parameter, minimum, maximum, what, desc):

        # Try to transform it to float, if it works than we transform it to a parameter

        try:

            number_or_parameter = float(number_or_parameter)

        except TypeError:

            assert isinstance(number_or_parameter, Parameter), "%s must be either a number or a " \
                                                               "parameter instance" % what

            # So this is a Parameter instance already. Enforce that it has the right maximum and minimum

            parameter = number_or_parameter

            assert parameter.min_value == minimum, "%s must have a minimum of %s" % (what, minimum)
            assert parameter.max_value == maximum, "%s must have a maximum of %s" % (what, maximum)

        else:

            # This was a float. Enforce that it has a legal value

            assert minimum <= number_or_parameter <= maximum, "%s cannot have a value of %s, " \
                                                              "it must be %s <= %s <= %s" % (what, number_or_parameter,
                                                                                             minimum, what, maximum)

            parameter = Parameter(what, number_or_parameter,
                                  desc=desc, min_value=minimum, max_value=maximum, unit='deg', free=False)

        return parameter

    def get_ra(self):
        """
        Get R.A. corresponding to the current position (ICRS, J2000)

        :return: Right Ascension
        """

        try:

            return self.ra.value

        except AttributeError:

            # Transform from L,B to R.A., Dec

            return self.sky_coord.transform_to('icrs').ra.value

    def get_dec(self):
        """
        Get Dec. corresponding to the current position (ICRS, J2000)

        :return: Declination
        """

        try:

            return self.dec.value

        except AttributeError:

            # Transform from L,B to R.A., Dec

            return self.sky_coord.transform_to('icrs').dec.value

    def get_l(self):
        """
        Get Galactic Longitude (l) corresponding to the current position

        :return: Galactic Longitude
        """

        try:

            return self.l.value

        except AttributeError:

            # Transform from L,B to R.A., Dec

            return self.sky_coord.transform_to('galactic').l.value

    def get_b(self):
        """
        Get Galactic latitude (b) corresponding to the current position

        :return: Latitude
        """

        try:

            return self.b.value

        except AttributeError:

            # Transform from L,B to R.A., Dec

            return self.sky_coord.transform_to('galactic').b.value

    def _get_sky_coord(self):

        if self._coord_type == 'galactic':

            l = self.l.value
            b = self.b.value

            return coordinates.SkyCoord(l=l, b=b,
                                        frame='galactic', equinox=self._equinox,
                                        unit="deg")

        else:

            ra = self.ra.value
            dec = self.dec.value

            return coordinates.SkyCoord(ra=ra, dec=dec,
                                        frame='icrs', equinox=self._equinox,
                                        unit="deg")

    @property
    def sky_coord(self):
        """
        Return an instance of astropy.coordinates.SkyCoord which can be used to make all transformations supported
        by it

        :return: astropy.coordinates.SkyCoord
        """
        return self._get_sky_coord()

    @property
    def parameters(self):
        """
        Get the dictionary of parameters (either ra,dec or l,b)

        :return: dictionary of parameters
        """

        if self._coord_type == 'galactic':

            return collections.OrderedDict((('l', self.l), ('b', self.b)))

        else:

            return collections.OrderedDict((('ra', self.ra), ('dec', self.dec)))

    @property
    def equinox(self):
        """
        Returns the equinox for the coordinates.

        :return:
        """
        return self._equinox

    def to_dict(self, minimal=False):

        data = collections.OrderedDict()

        if self._coord_type == 'equatorial':

            data['ra'] = self.ra.to_dict(minimal)
            data['dec'] = self.dec.to_dict(minimal)
            data['equinox'] = self._equinox

        else:

            data['l'] = self.l.to_dict(minimal)
            data['b'] = self.b.to_dict(minimal)
            data['equinox'] = self._equinox

        return data

    def fix(self):
        """
        Fix the parameters with the coordinates (either ra,dec or l,b depending on how the class
        has been instanced)
        
        """

        if self._coord_type == 'equatorial':

            self.ra.fix = True
            self.dec.fix = True

        else:

            self.l.fix = True
            self.b.fix = True

    def free(self):
        """
        Free the parameters with the coordinates (either ra,dec or l,b depending on how the class
        has been instanced)
        
        """

        if self._coord_type == 'equatorial':

            self.ra.fix = False
            self.dec.fix = False

        else:

            self.l.fix = False
            self.b.fix = False

    @classmethod
    def from_dict(cls, data):

        return cls(**data)

    def _repr__base(self, rich_output):

        if self._coord_type == 'equatorial':

            representation = 'Sky direction (R.A., Dec.) = (%.5f, %.5f) (%s)' % (self.ra.value,
                                                                                 self.dec.value,
                                                                                 self.equinox)

        else:

            representation = 'Sky direction (l, b) = (%.5f, %.5f) (%s)' % (self.l.value,
                                                                           self.b.value,
                                                                           self.equinox)

        return representation


from __future__ import absolute_import
# Import the version

#
#

import os

if os.environ.get("ASTROMODELS_DEBUG", None) is None:

    from .sources.point_source import PointSource
    from .sources.extended_source import ExtendedSource
    from .sources.particle_source import ParticleSource
    from .core.parameter import Parameter, IndependentVariable, SettingOutOfBounds

    from .functions.functions import *
    from .functions.priors import *
    from .functions.functions_2D import *
    from .functions.functions_3D import *
    from .functions.template_model import *
    from .functions.apec import *
    from .functions. dark_matter.dm_models import *
    from .functions.function import list_functions, get_function_class
    from .core.model import Model
    from .core.parameter import Parameter
    from .core.spectral_component import SpectralComponent
    from .core.polarization import LinearPolarization, StokesPolarization
    from .core.model_parser import load_model, clone_model
    from .core.units import get_units
    from .core.memoization import use_astromodels_memoization
    from .core.serialization import *

    astromodels_units = get_units()


import astropy.units as u

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import collections

import astropy.units as u
import numpy as np

from astromodels.core.spectral_component import SpectralComponent
from astromodels.core.tree import Node
from astromodels.core.units import get_units
from astromodels.functions.functions import Constant
from astromodels.sources.source import Source, EXTENDED_SOURCE
from astromodels.utils.pretty_list import dict_to_list


class ExtendedSource(Source, Node):

    def __init__(self, source_name, spatial_shape, spectral_shape=None, components=None):

        # Check that we have all the required information
        # and set the units

        current_u = get_units()

        if spatial_shape.n_dim == 2:

            # Now gather the component(s)

            # We need either a single component, or a list of components, but not both
            # (that's the ^ symbol)

            assert (spectral_shape is not None) ^ (components is not None), "You have to provide either a single " \
                                                                            "component, or a list of components " \
                                                                            "(but not both)."

            # If the user specified only one component, make a list of one element with a default name ("main")

            if spectral_shape is not None:

                components = [SpectralComponent("main", spectral_shape)]

            # Components in this case have energy as x and differential flux as y

            diff_flux_units = (current_u.energy * current_u.area * current_u.time) ** (-1)

            # Now set the units of the components
            for component in components:

                component.shape.set_units(current_u.energy, diff_flux_units)

            # Set the units of the brightness
            spatial_shape.set_units(current_u.angle, current_u.angle, current_u.angle**(-2))

        elif spatial_shape.n_dim == 3:

            # If there is no spectral component then assume that the input is a template, which will provide the
            # spectrum by itself. We just use a renormalization (a bias)

            if spectral_shape is None and components is None:

                # This is a template. Add a component which is just a renormalization

                spectral_shape = Constant()
                components = [SpectralComponent("main", spectral_shape)]

                # set the units
                diff_flux_units = (current_u.energy * current_u.area * current_u.time *
                                   current_u.angle**2) ** (-1)
                spatial_shape.set_units(current_u.angle, current_u.angle, current_u.energy, diff_flux_units)

            else:

                # the spectral shape has been given, so this is a case where the spatial template gives an
                # energy-dependent shape and the spectral components give the spectrum

                assert (spectral_shape is not None) ^ (components is not None), "You can provide either a single " \
                                                                                "component, or a list of components " \
                                                                                "(but not both)."

                if spectral_shape is not None:

                    components = [SpectralComponent("main", spectral_shape)]

                # Assign units
                diff_flux_units = (current_u.energy * current_u.area * current_u.time) ** (-1)

                # Now set the units of the components
                for component in components:
                    component.shape.set_units(current_u.energy, diff_flux_units)

                # Set the unit of the spatial template
                spatial_shape.set_units(current_u.angle, current_u.angle, current_u.energy, current_u.angle**(-2))

        else:

            raise RuntimeError("The spatial shape must have either 2 or 3 dimensions.")

        # Here we have a list of components

        Source.__init__(self, components, EXTENDED_SOURCE)

        # A source is also a Node in the tree

        Node.__init__(self, source_name)

        # Add the spatial shape as a child node, with an explicit name
        self._spatial_shape = spatial_shape
        self._add_child(self._spatial_shape)

        # Add the same node also with the name of the function
        #self._add_child(self._shape, self._shape.__name__)

        # Add a node called 'spectrum'

        spectrum_node = Node('spectrum')
        spectrum_node._add_children(list(self._components.values()))

        self._add_child(spectrum_node)

    @property
    def spatial_shape(self):
        """
        A generic name for the spatial shape.

        :return: the spatial shape instance
        """

        return self._spatial_shape

    def get_spatially_integrated_flux( self, energies):
    
        """
        Returns total flux of source at the given energy
        :param energies: energies (array or float)
        :return: differential flux at given energy
        """
       
        if not isinstance(energies, np.ndarray):
            energies = np.array(energies, ndmin=1)

        # Get the differential flux from the spectral components

        results = [self.spatial_shape.get_total_spatial_integral(energies) * component.shape(energies) for component in self.components.values()]

        if isinstance(energies, u.Quantity):

            # Slow version with units

            # We need to sum like this (slower) because using np.sum will not preserve the units
            # (thanks astropy.units)

            differential_flux = sum(results)

        else:

            # Fast version without units, where x is supposed to be in the same units as currently defined in
            # units.get_units()

            differential_flux = np.sum(results, 0)

        return differential_flux


    def __call__(self, lon, lat, energies):
        """
        Returns brightness of source at the given position and energy
        :param lon: longitude (array or float)
        :param lat: latitude (array or float)
        :param energies: energies (array or float)
        :return: differential flux at given position and energy
        """

        assert type(lat) == type(lon) and type(lon) == type(energies), "Type mismatch in input of call"

        if not isinstance(lat, np.ndarray):

            lat = np.array(lat, ndmin=1)
            lon = np.array(lon, ndmin=1)
            energies = np.array(energies, ndmin=1)

        # Get the differential flux from the spectral components

        results = [component.shape(energies) for component in list(self.components.values())]

        if isinstance(energies, u.Quantity):

            # Slow version with units

            # We need to sum like this (slower) because using np.sum will not preserve the units
            # (thanks astropy.units)

            differential_flux = sum(results)

        else:

            # Fast version without units, where x is supposed to be in the same units as currently defined in
            # units.get_units()

            differential_flux = np.sum(results, 0)

        # Get brightness from spatial model

        if self._spatial_shape.n_dim == 2:

            brightness = self._spatial_shape(lon, lat)

            # In this case the spectrum is the same everywhere
            n_points = lat.shape[0]
            n_energies = differential_flux.shape[0]

            # The following is a little obscure, but it is 6x faster than doing a for loop

            cube = np.repeat(differential_flux, n_points).reshape(n_energies, n_points).T
            result = (cube.T * brightness).T

        else:

            result = self._spatial_shape(lon, lat, energies) * differential_flux

        # Do not clip the output, otherwise it will not be possible to use ext. sources
        # with negative fluxes

        return np.squeeze(result)
              
    def has_free_parameters(self):
        """
        Returns True or False whether there is any parameter in this source

        :return:
        """

        for component in list(self._components.values()):

            for par in list(component.shape.parameters.values()):

                if par.free:

                    return True

        for par in list(self.spatial_shape.parameters.values()):

            if par.free:

                return True

        return False

    @property
    def free_parameters(self):
        """
        Returns a dictionary of free parameters for this source
        We use the parameter path as the key because it's 
        guaranteed to be unique, unlike the parameter name.

        :return:
        """
        free_parameters = collections.OrderedDict()

        for component in list(self._components.values()):

            for par in list(component.shape.parameters.values()):

                if par.free:

                    free_parameters[par.path] = par

        for par in list(self.spatial_shape.parameters.values()):

            if par.free:

                free_parameters[par.path] = par

        return free_parameters

    @property
    def parameters(self):
        """
        Returns a dictionary of all parameters for this source.
        We use the parameter path as the key because it's 
        guaranteed to be unique, unlike the parameter name.

        :return:
        """
        all_parameters = collections.OrderedDict()

        for component in list(self._components.values()):

            for par in list(component.shape.parameters.values()):

                all_parameters[par.path] = par

        for par in list(self.spatial_shape.parameters.values()):

            all_parameters[par.path] = par

        return all_parameters

    def _repr__base(self, rich_output=False):
        """
        Representation of the object

        :param rich_output: if True, generates HTML, otherwise text
        :return: the representation
        """

        # Make a dictionary which will then be transformed in a list

        repr_dict = collections.OrderedDict()

        key = '%s (extended source)' % self.name

        repr_dict[key] = collections.OrderedDict()
        repr_dict[key]['shape'] = self._spatial_shape.to_dict(minimal=True)
        repr_dict[key]['spectrum'] = collections.OrderedDict()

        for component_name, component in list(self.components.items()):
            repr_dict[key]['spectrum'][component_name] = component.to_dict(minimal=True)

        return dict_to_list(repr_dict, rich_output)

    def get_boundaries(self):
        """
        Returns the boundaries for this extended source

        :return: a tuple of tuples ((min. lon, max. lon), (min lat, max lat))
        """
        return self._spatial_shape.get_boundaries()



from __future__ import print_function
from builtins import range
import pytest
import os


os.environ["ASTROMODELS_DEBUG"] = "debug"

from astromodels.core.tree import Node
from astromodels.core import node_ctype

import astropy.units as u

import gc

def clean():

    gc.collect()

class _SimpleInheritance(Node):

    def __init__(self, name):
        print("INIT of _SimpleInheritance")

        self._placeholder = 2

        super(_SimpleInheritance, self).__init__(name)

    @property
    def placeholder(self):
        return self._placeholder


class _ComplexInheritance(Node):

    def __init__(self, name, min_value, max_value):

        super(_ComplexInheritance, self).__init__(name)

        self._min_value = min_value * u.Unit("1 / (keV * cm**2 * s)")
        self._max_value = max_value * u.Unit("1 / (keV * cm**2 * s)")

    @property
    def min_value(self):

        return self._min_value

    @property
    def max_value(self):

        return self._max_value


def test_constructor():

    n = Node(name='node1')

    assert n.name == 'node1'

    with pytest.raises(TypeError):

        n2 = Node()

    clean()


def test_inheritance():
    t = _SimpleInheritance('test1')

    assert t.name == 'test1'

    clean()


def test_add_child():

    t = _SimpleInheritance('test_add_child')

    with pytest.raises(TypeError):
        t._add_child("clara")

    clara = Node("clara")

    clara_ref_count = node_ctype._get_reference_counts(clara)

    t._add_child(clara)

    # We expect 2 more references: one for the map, one for the vector

    assert node_ctype._get_reference_counts(clara) == clara_ref_count + 1

    assert t.clara == clara

    with pytest.raises(AttributeError):
        t.clara = 'argh'

    assert t.clara == clara

    clean()


def test_add_children():
    t = _SimpleInheritance("children")

    t._add_children([Node("node1"), Node("node2")])

    clean()


def test_get_child():
    t = _SimpleInheritance("test")

    n = Node("node")

    t._add_child(n)

    assert t._get_child("node") == n

    with pytest.raises(AttributeError):
        t._get_child("not existing")

    clean()


def test_get_children():
    node1 = Node('node1')
    node2 = Node('node2')
    node22 = Node('node22')
    node3 = Node('node3')

    node1._add_child(node2)
    node1._add_child(node22)
    node2._add_child(node3)

    assert node1._get_children() == (node2, node22)
    assert node2._get_children() == (node3,)

    clean()


def test_remove_child():
    t = _SimpleInheritance("test")

    n = Node("node")

    for i in range(1000):

        t._add_child(n)

        assert t._get_child("node") == n

        t._remove_child("node")

    with pytest.raises(AttributeError):
        print(t.node)

    with pytest.raises(AttributeError):
        t._get_child("node")

    clean()


def test_get_path():
    # Make a small tree

    node1 = Node('node1')
    node2 = Node('node2')
    node3 = Node('node3')

    node1._add_child(node2)
    node2._add_child(node3)

    print(node3.path)

    assert node3.path == "node1.node2.node3"

    clean()


def test_get_child_from_path():
    # Make a small tree

    node1 = Node('node1')
    node2 = Node('node2')
    node3 = Node('node3')

    node1._add_child(node2)
    node2._add_child(node3)

    assert node1._get_child_from_path("node2.node3") == node3

    clean()


def test_change_name():
    t = _SimpleInheritance("name1")

    assert t.name == "name1"

    with pytest.raises(AttributeError):
        t.name = "name2"

    assert t.name == "name1"

    t._change_name("name2")

    assert t.name == "name2"

    clean()


def test_pickle():

    print("\n\n\n\n")

    import pickle

    root = Node("root")
    node = Node("node")

    root._add_child(node)

    d = pickle.dumps(root)

    root2 = pickle.loads(d)

    assert root2.node.path == 'root.node'
    assert root2.node.name == 'node'

    # Now test pickling a subclass of Node

    root = _SimpleInheritance("root")

    root._placeholder = 5.3

    node = _SimpleInheritance("node")

    root._add_child(node)

    d = pickle.dumps(root)

    root2 = pickle.loads(d)

    assert root2.node.path == 'root.node'
    assert root2.node.name == 'node'

    print(root.placeholder)

    assert root2.placeholder == root.placeholder

    rootc = _ComplexInheritance("root", -1.0, 1.0)
    nodec = _ComplexInheritance("node", -5.0, 5.0)

    rootc._add_child(nodec)

    d = pickle.dumps(rootc)

    root2c = pickle.loads(d)

    print(root2c.min_value)

    clean()


def test_memory_leaks_setters():

    root = Node("root")

    node = Node("node")

    refc_before_link = node_ctype._get_reference_counts(node)

    root._add_child(node)

    # Adding a node adds 1 reference

    assert node_ctype._get_reference_counts(node) == refc_before_link + 1

    # Remove the node and verify that the reference counts goes back to what it was
    root._remove_child("node")

    # Now we should go back to the original

    assert node_ctype._get_reference_counts(node) == refc_before_link

    # Now add a second node nested under the first one (root.node.node2)
    node2 = Node("node2")

    refc_before_link2 = node_ctype._get_reference_counts(node2)

    root._add_child(node)  # +1 for node

    assert node_ctype._get_reference_counts(node) == refc_before_link + 1

    node._add_child(node2)  # +1 for node2 and +1 for node

    assert node_ctype._get_reference_counts(node2) == refc_before_link2 + 1

    # "node" is now also parent of node2, so its reference are now 2 more than the original
    assert node_ctype._get_reference_counts(node) == refc_before_link + 2

    # Clean up and verify that we go back to the original number of references
    node._remove_child("node2")  # -1 for node2, -1 for node

    assert node_ctype._get_reference_counts(node2) == refc_before_link2

    assert node_ctype._get_reference_counts(node) == refc_before_link + 1

    root._remove_child("node")  # -1 for node

    assert node_ctype._get_reference_counts(node) == refc_before_link

    # Now test add_children
    node3 = Node("node3")
    node4 = Node("node4")

    refc_before_link3 = node_ctype._get_reference_counts(node3)
    refc_before_link4 = node_ctype._get_reference_counts(node4)

    root._add_children([node3, node4])  # +1 for both nodes

    assert node_ctype._get_reference_counts(node3) == refc_before_link3 + 1
    assert node_ctype._get_reference_counts(node3) == refc_before_link4 + 1


def test_memory_leaks_getters():

    # Now test the getters

    root = Node("root")

    node1 = Node("node1")
    node2 = Node("node2")
    node3 = Node("node3")

    refc_before_link_root = node_ctype._get_reference_counts(root)
    refc_before_link1 = node_ctype._get_reference_counts(node1)
    refc_before_link2 = node_ctype._get_reference_counts(node2)
    refc_before_link3 = node_ctype._get_reference_counts(node3)

    None_counts_before = node_ctype._get_reference_counts(None)

    # Add 3 nodes to root
    root._add_children([node1, node2, node3]) # node1 +1, node2 +1, node3 + 1, root +3

    node_again = root._get_child("node1")  # node1 +1

    assert node_ctype._get_reference_counts(node1) == refc_before_link1 + 2

    del node_again # node1 -1

    assert node_ctype._get_reference_counts(node1) == refc_before_link1 + 1

    children = root._get_children()  #node1 +1, node2 +1, node3 +1

    assert len(children) == 3

    assert node_ctype._get_reference_counts(node1) == refc_before_link1 + 2
    assert node_ctype._get_reference_counts(node2) == refc_before_link2 + 2
    assert node_ctype._get_reference_counts(node3) == refc_before_link3 + 2
    assert node_ctype._get_reference_counts(root) == refc_before_link_root + 3

    # test get_parent

    root_again = node1._get_parent()  # root +1

    assert node_ctype._get_reference_counts(root) == refc_before_link_root + 4

    del root_again  # root -1

    assert node_ctype._get_reference_counts(root) == refc_before_link_root + 3

    # test _get_path

    path = node2._get_path()  # this shouldn't change any ref count

    assert node_ctype._get_reference_counts(node1) == refc_before_link1 + 2
    assert node_ctype._get_reference_counts(node2) == refc_before_link2 + 2
    assert node_ctype._get_reference_counts(node3) == refc_before_link3 + 2
    assert node_ctype._get_reference_counts(root) == refc_before_link_root + 3

    # test _get_child_from_path

    node4 = Node("node4")
    refc_before_link4 = node_ctype._get_reference_counts(node4)

    node3._add_child(node4)  # node3 +1, node4 + 1

    node4_again = root._get_child_from_path("node3.node4")  # node4 +1

    assert node_ctype._get_reference_counts(node4) == refc_before_link4 + 2
    assert node_ctype._get_reference_counts(node3) == refc_before_link3 + 3

    del node4_again  # node4 -1

    assert node_ctype._get_reference_counts(node4) == refc_before_link4 + 1


def test_memory_leaks_destructors():

    print("\n\n\n\n\n")

    for i in range(1000):

        print("\n\nRound %i" % (i+1))

        # Test destructors
        root = Node("root")

        node1 = Node("node1")
        node2 = Node("node2")
        node3 = Node("node3")

        refc_before_link_root = node_ctype._get_reference_counts(root)

        refc_before_link1 = node_ctype._get_reference_counts(node1)
        refc_before_link2 = node_ctype._get_reference_counts(node2)
        refc_before_link3 = node_ctype._get_reference_counts(node3)

        # root._add_children([node1, node2, node3])  # node1 +1, node2 + 1, node3 + 1, root +3

        root._add_child(node1)
        root._add_child(node2)
        root._add_child(node3)


        assert node_ctype._get_reference_counts(root) == refc_before_link_root + 3
        assert node_ctype._get_reference_counts(node1) == refc_before_link1 + 1
        assert node_ctype._get_reference_counts(node2) == refc_before_link2 + 1
        assert node_ctype._get_reference_counts(node3) == refc_before_link3 + 1

        #
        # Let's destroy the tree
        root._remove_child("node1")

        root._remove_child("node2")

        root._remove_child("node3")
        #
        # assert node_ctype._get_reference_counts(root) == refc_before_link_root + 3
        #
        # # Now deleting root should have decreased all nodes by 3, i..e, should have got them back to the initial
        # # reference counts
        # assert node_ctype._get_reference_counts(node1) == refc_before_link1
        # assert node_ctype._get_reference_counts(node2) == refc_before_link2
        # assert node_ctype._get_reference_counts(node3) == refc_before_link3



from __future__ import print_function
from __future__ import division
from builtins import object
import pytest
import os

import astropy.units as u
import numpy as np
import pickle

from astromodels.functions.function import FunctionMeta, Function1D, Function2D, FunctionDefinitionError, \
    UnknownParameter, DesignViolation, get_function, get_function_class, UnknownFunction, list_functions
from astromodels.functions.functions import Powerlaw, Line
from astromodels.functions.functions_2D import Gaussian_on_sphere, SpatialTemplate_2D
from astromodels.functions.functions_3D import Continuous_injection_diffusion
from astromodels.functions import function as function_module

from astropy.io import fits
from future.utils import with_metaclass

__author__ = 'giacomov'


def get_a_function_class():

    # Try to create a function inheriting from Function with meta FunctionMeta
    class Test_function(with_metaclass(FunctionMeta, Function1D)):
        r"""
        description :

            A test function

        latex : $ a * x + b $

        parameters :

            a :

                desc : linear coefficient
                initial value : 1

            b :

                desc : intercept
                initial value : 1

        """

        def _set_units(self, x_unit, y_unit):

            # a has units of y_unit / x_unit, so that a*x has units of y_unit
            self.a.unit = y_unit / x_unit

            # b has units of y
            self.b.unit = y_unit

        def evaluate(self, x, a, b):

            return a * x + b

    return Test_function


def test_function_meta():

    with pytest.raises(AttributeError):

        # .evaluate is lacking, ._set_units is lacking, docstring is lacking

        class Wrong_test_function1(with_metaclass(FunctionMeta, object)):
            pass

    with pytest.raises(AttributeError):

        # .evaluate is lacking, ._set_units is lacking

        class Wrong_test_function2(with_metaclass(FunctionMeta, Function1D)):
            r"""
            description :

                A test function

            latex : $ a * x + b $

            parameters :

                a :

                    desc : linear coefficient
                    initial value : 1

                b :

                    desc : intercept
                    initial value : 1

            """

    with pytest.raises(AttributeError):
        # _set_units is lacking

        class Wrong_test_function3(with_metaclass(FunctionMeta, Function1D)):
            r"""
            description :

                A test function

            latex : $ a * x + b $

            parameters :

                a :

                    desc : linear coefficient
                    initial value : 1

                b :

                    desc : intercept
                    initial value : 1

            """

            def evaluate(self, x, a, b):

                return a * x + b

    with pytest.raises(AssertionError):
        # Signature of evaluate is wrong

        class Wrong_test_function4(with_metaclass(FunctionMeta, Function1D)):
            r"""
            description :

                A test function

            latex : $ a * x + b $

            parameters :

                a :

                    desc : linear coefficient
                    initial value : 1

                b :

                    desc : intercept
                    initial value : 1

            """

            def _set_units(self, x_unit, y_unit):

                self.a.unit = y_unit / x_unit
                self.b.unit = y_unit

            def evaluate(self):

                return self.a * self.x + self.b

    with pytest.raises(FunctionDefinitionError):
        # Signature of evaluate is wrong

        class Wrong_test_function5(with_metaclass(FunctionMeta, Function1D)):
            r"""
            description :

                A test function

            latex : $ a * x + b $

            parameters :

                a :

                    desc : linear coefficient
                    initial value : 1

                b :

                    desc : intercept
                    initial value : 1

            """

            def _set_units(self, x_unit, y_unit):
                self.a.unit = y_unit / x_unit
                self.b.unit = y_unit

            def evaluate(self, x):
                return self.a * x + self.b

    with pytest.raises(FunctionDefinitionError):
        # Signature of evaluate is wrong

        class Wrong_test_function6(with_metaclass(FunctionMeta, Function1D)):
            r"""
            description :

                A test function

            latex : $ a * x + b $

            parameters :

                a :

                    desc : linear coefficient
                    initial value : 1

                b :

                    desc : intercept
                    initial value : 1

            """

            def _set_units(self, x_unit, y_unit):
                self.a.unit = y_unit / x_unit
                self.b.unit = y_unit

            def evaluate(self, x, a):
                return a * x + self.b

    with pytest.raises(FunctionDefinitionError):
        # Signature of evaluate does not match docstring

        class Wrong_test_function7(with_metaclass(FunctionMeta, Function1D)):
            r"""
            description :

                A test function

            latex : $ a * x + b $

            parameters :

                a :

                    desc : linear coefficient
                    initial value : 1

                c :

                    desc : intercept
                    initial value : 1

            """

            def _set_units(self, x_unit, y_unit):
                self.a.unit = y_unit / x_unit
                self.b.unit = y_unit

            def evaluate(self, x, a, b):
                return a * x + b

    with pytest.raises(FunctionDefinitionError):
        # Definition of parameter b is not legal

        class Wrong_test_function8(with_metaclass(FunctionMeta, Function1D)):
            r"""
            description :

                A test function

            latex : $ a * x + b $

            parameters :

                a :

                    desc : linear coefficient
                    initial value : 1

                b :

                    desc : intercept

            """

            def _set_units(self, x_unit, y_unit):
                self.a.unit = y_unit / x_unit
                self.b.unit = y_unit

            def evaluate(self, x, a, b):
                return a * x + b

    with pytest.raises(FunctionDefinitionError):
        # Parameter c declared but not used

        class Wrong_test_function9(with_metaclass(FunctionMeta, Function1D)):
            r"""
            description :

                A test function

            latex : $ a * x + b $

            parameters :

                a :

                    desc : linear coefficient
                    initial value : 1

                b :

                    desc : intercept
                    initial value : 1

                c :

                    desc : dumb
                    initial value : 1

            """

            def _set_units(self, x_unit, y_unit):
                self.a.unit = y_unit / x_unit
                self.b.unit = y_unit

            def evaluate(self, x, a, b):
                return a * x + b

    with pytest.raises(FunctionDefinitionError):
        # Parameter c used but not declared

        class Wrong_test_function10(with_metaclass(FunctionMeta, Function1D)):
            r"""
            description :

                A test function

            latex : $ a * x + b $

            parameters :

                a :

                    desc : linear coefficient
                    initial value : 1

                b :

                    desc : intercept
                    initial value : 1

            """

            def _set_units(self, x_unit, y_unit):
                self.a.unit = y_unit / x_unit
                self.b.unit = y_unit

            def evaluate(self, x, a, b, c):
                return a * x + b + c


    with pytest.raises(AssertionError):
        # Docstring lacking description

        class Wrong_test_function11(with_metaclass(FunctionMeta, Function1D)):
            r"""
            latex : $ a * x + b $

            parameters :

                a :

                    desc : linear coefficient
                    initial value : 1

                b :

                    desc : intercept
                    initial value : 1

                c :

                    desc : dumb
                    initial value : 1

            """

            def _set_units(self, x_unit, y_unit):
                self.a.unit = y_unit / x_unit
                self.b.unit = y_unit

            def evaluate(self, x, a, b):
                return a * x + b

    with pytest.raises(FunctionDefinitionError):
        # Parameter lacking description

        class Wrong_test_function12(with_metaclass(FunctionMeta, Function1D)):
            r"""

            description: useless

            latex : $ a * x + b $

            parameters :

                a :

                    initial value : 1

                b :

                    desc : intercept
                    initial value : 1

                c :

                    desc : dumb
                    initial value : 1

            """

            def _set_units(self, x_unit, y_unit):
                self.a.unit = y_unit / x_unit
                self.b.unit = y_unit

            def evaluate(self, x, a, b):
                return a * x + b

    with pytest.raises(AssertionError):
        # Parameters out of order in evaluate

        class Wrong_test_function13(with_metaclass(FunctionMeta, Function2D)):
            r"""

            description: useless

            latex : $ a * x + b $

            parameters :

                a :

                    desc : blah
                    initial value : 1


                b :

                    desc : intercept
                    initial value : 1

                c :

                    desc : dumb
                    initial value : 1

            """

            def _set_units(self, x_unit, y_unit, z_unit):
                self.a.unit = y_unit / x_unit
                self.b.unit = y_unit

            def evaluate(self, y, x, a, b, c):
                return a * x + b

    # A function with no latex formula (which is optional)

    class NoLatex_test_function11(with_metaclass(FunctionMeta, Function1D)):
        r"""

        description:

            A function without latex (should be fine)

        parameters :

            a :

                desc : linear coefficient
                initial value : 1

            b :

                desc : intercept
                initial value : 1

        """

        def _set_units(self, x_unit, y_unit):
            self.a.unit = y_unit / x_unit
            self.b.unit = y_unit

        def evaluate(self, x, a, b):
            return a * x + b


def test_function_constructor():

    Test_function = get_a_function_class()

    # Instance with default parameters
    my_function = Test_function()

    assert my_function.a.value == 1.0
    assert my_function.b.value == 1.0

    # Instance with explicit parameters' values
    my_function = Test_function(b=3.2, a=-2.5)

    assert my_function.a.value == -2.5
    assert my_function.b.value == 3.2

    Test_function.info()

    function_module.has_ipython = False

    Test_function.info()

    print(my_function.free_parameters)

    with pytest.raises(UnknownParameter):

        f = Test_function(d=3.5)

    f = Test_function()

    print(f.description)

    print(f.latex)

    assert f.fixed_units is None

    assert f.has_fixed_units() == False

    with pytest.raises(DesignViolation):

        _ = f.get_boundaries()


def test_function_values():

    # Get a function class, and test the various exceptions

    Test_function = get_a_function_class()

    # Try to instance it
    my_function = Test_function()

    # Test basic functionalities

    assert my_function(1.0)==2

    my_function.a = 2.5

    assert my_function(10.0) == 26.0

    my_function.b = -1.0

    assert my_function(10.0) == 24.0

    # Now test with list and np.array

    my_function.a.value = 2.0
    my_function.b.value = 1.0

    assert np.all(my_function([1,2,3]) == np.array([3.0, 5.0, 7.0]))
    assert np.all(my_function(np.array([3, 4, 5])) == np.array([7.0, 9.0, 11.0]))


def test_function_values_units():

    # Test units functionality
    # Get a function class, and test the various exceptions

    Test_function = get_a_function_class()

    # Try to instance it
    my_function = Test_function()

    diff_flux = 1.0 / (u.keV * u.cm**2 * u.s)

    my_function.set_units(u.keV, diff_flux)

    # Test basic functionalities

    assert my_function(1.0 * u.keV) == 2 * diff_flux

    my_function.a = 2.5 * diff_flux / u.keV

    assert my_function(10.0 * u.keV) == 26.0 * diff_flux

    my_function.b = -1.0 * diff_flux

    assert my_function(10.0 * u.keV) == 24.0 * diff_flux

    # Now test with list and np.array

    my_function.a.value = 2.0 * diff_flux / u.keV
    my_function.b.value = 1.0 * diff_flux

    assert np.all(my_function([1, 2, 3] * u.keV) == np.array([3.0, 5.0, 7.0]) * diff_flux)

    # Using one unit for each element will fail

    # (depending on the version of astropy, it might raise ValueError or TypeError)
    with pytest.raises((ValueError, TypeError)):

        _ = my_function([1 * u.keV, 2 * u.keV, 3 * u.keV])

    assert np.all(my_function(np.array([3, 4, 5]) * u.keV) == np.array([7.0, 9.0, 11.0]) * diff_flux)

    # Now test that an error is raised if units are not intelligible
    with pytest.raises(TypeError):

        _ = my_function.set_units("non_existent","non_existent")


def test_function_composition():

    Test_function = get_a_function_class()

    line = Test_function()
    powerlaw = Powerlaw()

    composite = powerlaw + line

    composite.set_units(u.keV, 1.0 / (u.keV * u.cm**2 * u.s))

    for x in ([1,2,3,4],[1,2,3,4] * u.keV, 1.0, np.array([1.0, 2.0, 3.0, 4.0])):

        assert np.all(composite(x) == line(x) + powerlaw(x))

    # Test -
    po = Powerlaw()
    li = Line()
    composite = po - li

    assert composite(1.0) == (po(1.0) - li(1.0))

    # test *
    composite = po * li

    assert composite(2.25) == po(2.25) * li(2.25)

    # test /
    composite = po / li

    assert composite(2.25) == po(2.25) / li(2.25)

    # test .of
    composite = po.of(li)

    assert composite(2.25) == po(li(2.25))

    # test power
    composite = po**li

    assert composite(2.25)  == po(2.25)**li(2.25)

    # test negation
    neg_po = -po

    assert neg_po(2.25) == -po(2.25)

    # test abs
    new_li = Line()
    new_li.b = -10.0

    abs_new_li = abs(new_li)

    assert new_li(1.0) < 0
    assert abs_new_li(1.0) == abs(new_li(1.0))

    # test rpower
    composite = 2.0**new_li

    assert composite(2.25) == 2.0**(new_li(2.25))

    # test multiplication by a number
    composite = 2.0 * po

    assert composite(2.25) == 2.0 * po(2.25)

    # Number divided by
    composite = 1.0 / li

    assert composite(2.25) == 1.0 / li(2.25)

    # Composite of composite
    composite = po*li + po - li + 2*po / li

    assert composite(2.25) == po(2.25) * li(2.25) + po(2.25) - li(2.25) + 2*po(2.25) / li(2.25)

    print(composite)


def test_duplicate():

    instance = Powerlaw()
    instance.index = -2.25
    instance.K = 0.5

    # Duplicate it

    duplicate = instance.duplicate()

    # Check that we have the same results

    assert duplicate(2.25) == instance(2.25)

    # Check that the parameters are not linked anymore
    instance.index = -1.12

    assert instance.index.value != duplicate.index.value

    print(instance)
    print(duplicate)


def test_pickling_unpickling():

    # 1d function
    po = Powerlaw()

    po.K = 5.35

    new_po = pickle.loads(pickle.dumps(po))

    assert new_po.K.value == po.K.value

    # 2d function
    gs = Gaussian_on_sphere()

    _ = pickle.loads(pickle.dumps(gs))

    # 3d function
    c = Continuous_injection_diffusion()

    _ = pickle.loads(pickle.dumps(c))

    # composite function
    po2 = Powerlaw()
    li = Line()
    composite = po2*li + po2 - li + 2*po2 / li  # type: Function1D

    # Change some parameter
    composite.K_1 = 3.2
    composite.a_2 = 1.56

    dump = pickle.dumps(composite)

    new_composite = pickle.loads(dump)

    assert new_composite.K_1.value == composite.K_1.value
    assert new_composite.a_2.value == composite.a_2.value


def test_get_function():

    po = get_function("Powerlaw")

    _ = po(1.0)

    with pytest.raises(UnknownFunction):

        _ = get_function("not_existant")


def test_get_function_class():

    po_class = get_function_class("Powerlaw")

    assert po_class == Powerlaw

    with pytest.raises(UnknownFunction):

        _ = get_function_class("not_existant")


def test_list_functions():

    print(list_functions())


def test_function2D():

    c = Gaussian_on_sphere()

    f1 = c(1, 1)
    assert np.isclose( f1, 38.285617800653434, rtol=1e-10)
    

    a = np.array([1.0, 2.0])

    fa = c(a, a)
    assert np.isclose( fa, [3.82856178e+01, 2.35952748e-04], rtol=1e-10).all()

    c.set_units(u.deg, u.deg, 1.0 / u.deg**2)

    f1d = c(1 * u.deg, 1.0 * u.deg)
    assert np.isclose( f1d.value, 38.285617800653434, rtol=1e-10)
    assert f1d.unit == u.deg**-2

    assert c.x_unit == u.deg
    assert c.y_unit == u.deg
    assert c.z_unit == u.deg**-2

    assert c.get_total_spatial_integral( 1 ) == 1
    assert np.isclose( c.get_total_spatial_integral( [1,1] ) ,  [1,1], rtol=1e-10).all()
    

    with pytest.raises(TypeError):

        c.set_units("not existent", u.deg, u.keV)



def test_function3D():

    c = Continuous_injection_diffusion()

    f1 = c(1, 1, 1)
    assert np.isclose(f1, 134.95394313247866, rtol = 1e-10)
    
    a = np.array([1.0, 2.0])

    fa = c(a, a, a)
    assert np.isclose( fa,  [[134.95394313, 132.19796573], [ 25.40751507, 27.321443  ]], rtol=1e-10).all()

    c.set_units(u.deg, u.deg, u.keV, 1.0 / u.deg**2)

    f1d = c(1 * u.deg, 1.0 * u.deg, 1.0 * u.keV)
    assert np.isclose(f1d.value, 134.95394313247866, rtol = 1e-10)
    assert f1d.unit == u.deg**-2
   

    assert c.x_unit == u.deg
    assert c.y_unit == u.deg
    assert c.z_unit == u.keV
    assert c.w_unit == u.deg**-2

    assert c.get_total_spatial_integral( 1 ) == 1
    assert np.isclose( c.get_total_spatial_integral( [1,1] ) ,  [1,1], rtol=1e-10).all()

    with pytest.raises(TypeError):

        c.set_units("not existent", u.deg, u.keV, 1.0 / (u.keV * u.s * u.deg**2 * u.cm**2))


def test_spatial_template_2D():

    #make the fits files with templates to test.
    cards = {
      "SIMPLE": "T",                     
      "BITPIX": -32,
      "NAXIS" : 2,
      "NAXIS1": 360,
      "NAXIS2": 360,
      "DATE": '2018-06-15',  
      "CUNIT1": 'deg', 
      "CRVAL1":  83,
      "CRPIX1": 0,
      "CDELT1": -0.0166667, 
      "CUNIT2": 'deg',
      "CRVAL2": -2.0,
      "CRPIX2": 0,
      "CDELT2": 0.0166667,
      "CTYPE1": 'GLON-CAR',
      "CTYPE2": 'GLAT-CAR' }

    data = np.zeros([400,400])
    data[0:100,0:100] = 1
    hdu = fits.PrimaryHDU(data=data, header=fits.Header(cards))
    hdu.writeto("test1.fits", overwrite=True)

    data[:,:]=0
    data[200:300,200:300] = 1
    hdu = fits.PrimaryHDU(data=data, header=fits.Header(cards))
    hdu.writeto("test2.fits", overwrite=True)


    #Now load template files and test their evaluation
    shape1=SpatialTemplate_2D()
    shape1.load_file("test1.fits")
    shape1.K = 1

    shape2=SpatialTemplate_2D()
    shape2.load_file("test2.fits")
    shape2.K = 1

    assert shape1.hash != shape2.hash
        
    assert np.all ( shape1.evaluate( [312, 306], [41, 41], [1,1], [40, 2]) == [1., 0.] ) 
    assert np.all ( shape2.evaluate( [312, 306], [41, 41], [1,1], [40, 2]) ==  [0., 1.] ) 
    assert np.all ( shape1.evaluate( [312, 306], [41, 41], [1,10], [40, 2]) == [1., 0.] ) 
    assert np.all ( shape2.evaluate( [312, 306], [41, 41], [1,10], [40, 2]) ==  [0., 10.] ) 


    shape1.K = 1
    shape2.K = 1
    assert np.all ( shape1( [312, 306], [41, 41]) == [1., 0.] )
    assert np.all ( shape2( [312, 306], [41, 41]) == [0., 1.] )

    shape1.K = 1
    shape2.K = 10
    assert np.all ( shape1( [312, 306], [41, 41]) == [1., 0.] )
    assert np.all ( shape2( [312, 306], [41, 41]) == [0., 10.] )

    os.remove("test1.fits")
    os.remove("test2.fits")






# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

from __future__ import absolute_import
import json

version_json = '''
{
 "date": "2020-10-15T14:48:01-0700",
 "dirty": false,
 "error": null,
 "full-revisionid": "14c6a91c0696d02324452c6a02ce5db440d032b5",
 "version": "2.0.0"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



from __future__ import print_function
import astropy.units as u
import numpy as np
import pytest

from astromodels.core.spectral_component import SpectralComponent
from astromodels.functions.functions import Powerlaw, Log_parabola
from astromodels.functions.functions_2D import *
from astromodels.functions.functions_3D import *
from astromodels.sources.extended_source import ExtendedSource
from astromodels.core.model import Model
from astromodels.core.model_parser import clone_model


from astromodels.functions.priors import *
from astromodels.functions.function import _known_functions

__author__ = 'henrikef'

def make_test_template( ra, dec, fitsfile):

#Test template function: 40 pixel (0.8 deg) wide square centered approximately around a given ra, dec.

  cards = {
    "SIMPLE": "T",                     
    "BITPIX": -32,
    "NAXIS" : 2,
    "NAXIS1": 200,
    "NAXIS2": 200,
    "DATE": '2018-11-13',  
    "CUNIT1": 'deg', 
    "CRVAL1":  ra,
    "CRPIX1": 100,
    "CDELT1": -0.02, 
    "CUNIT2": 'deg',
    "CRVAL2": dec,
    "CRPIX2": 100,
    "CDELT2": 0.02,
    "CTYPE1": 'RA---TAN',
    "CTYPE2": 'DEC--TAN' }

  data = np.zeros([200,200])
  data[80:120,80:120] = 1
  
  total = np.sum(data)
  dOmega = (abs( cards["CDELT1"] * cards["CDELT2"])*u.degree*u.degree).to(u.steradian).value
  data=data/ total / dOmega

  hdu = fits.PrimaryHDU(data=data, header=fits.Header(cards))
  hdu.writeto(fitsfile, overwrite=True)

def test_constructor():

    # RA, Dec and L,B of the same point in the sky

    ra, dec = (125.6, -75.3)
    l, b = (288.44190139183564, -20.717313145391525)

    # This should throw an error as we are using Powerlaw instead of Powerlaw()
    with pytest.raises(RuntimeError):

        _ = ExtendedSource("my_source", Gaussian_on_sphere, Powerlaw)

    # This should throw an error because we should use a 2D function for the spatial shape
    with pytest.raises(RuntimeError):

        _ = ExtendedSource("my_source", Powerlaw(), Powerlaw())

    # Init with RA, Dec

    shape = Gaussian_on_sphere()
    source1 = ExtendedSource('my_source', shape, Powerlaw())
    shape.lon0=ra*u.degree
    shape.lat0=dec*u.degree

    assert source1.spatial_shape.lon0.value == ra
    assert source1.spatial_shape.lat0.value == dec

    # Verify that the position is free by default
    assert source1.spatial_shape.lon0.free
    assert source1.spatial_shape.lon0.free



def test_call():

    # Multi-component

    po1 = Powerlaw()
    po2 = Powerlaw()

    c1 = SpectralComponent("component1", po1)
    c2 = SpectralComponent("component2", po2)

    ra, dec = (125.6, -75.3)

    def test_one(class_type, name ):
        
      print("testing %s ..." % name)

      shape = class_type()
      source = ExtendedSource('test_source_%s' % name, shape, components=[c1, c2])
      
      if name != "SpatialTemplate_2D":
        shape.lon0=ra*u.degree
        shape.lat0=dec*u.degree
        
      else:
        make_test_template(ra, dec, "__test.fits")
        shape.load_file( "__test.fits" )
        shape.K = 1.0

      assert np.all(source.spectrum.component1([1, 2, 3]) == po1([1, 2, 3]))
      assert np.all(source.spectrum.component2([1, 2, 3]) == po2([1, 2, 3]))

      one = source.spectrum.component1([1, 2, 3])
      two = source.spectrum.component2([1, 2, 3])

      #check spectral components
      assert np.all( np.abs(one + two - source.get_spatially_integrated_flux([1,2,3])) == 0 )
    
      #check spectral and spatial components
      total = source( [ra, ra, ra], [dec, dec, dec], [1,2,3])
      spectrum = one + two
      spatial = source.spatial_shape( [ra, ra, ra], [dec, dec, dec] )
      assert np.all( np.abs( total - spectrum*spatial ) == 0 )
    
      total = source( [ra*1.01]*3, [dec*1.01]*3, [1,2,3])
      spectrum = one + two
      spatial = source.spatial_shape( [ra*1.01]*3, [dec*1.01]*3 )
      assert np.all( np.abs( total - spectrum*spatial ) == 0 )
    
    for key in _known_functions:

        if key in [ "Latitude_galactic_diffuse"]:
        #not testing latitude galactic diffuse for now.
            continue
        
        this_function = _known_functions[key]

        if this_function._n_dim == 2  and not this_function().is_prior:

            test_one(this_function, key)
            
    with pytest.raises(AssertionError):
        #this will fail because the Latitude_galactic_diffuse function isn't normalized.
        test_one(_known_functions["Latitude_galactic_diffuse"], "Latitude_galactic_diffuse")
   
 
def test_call_with_units():

    # Multi-component

    po1 = Powerlaw()
    po2 = Powerlaw()

    c1 = SpectralComponent("component1", po1)
    c2 = SpectralComponent("component2", po2)

    ra, dec = (125.6, -75.3)

    def test_one(class_type, name ):
        
      print("testing %s ..." % name)


      shape = class_type()
      source = ExtendedSource('test_source_%s' % name, spatial_shape = shape, components=[c1, c2])

      if name != "SpatialTemplate_2D":
        shape.lon0=ra*u.degree
        shape.lat0=dec*u.degree
        
      else:
        make_test_template(ra, dec, "__test.fits")
        shape.load_file( "__test.fits" )
        shape.K = 1.0

      assert np.all(source.spectrum.component1([1, 2, 3]*u.keV) == po1([1, 2, 3]*u.keV))
      assert np.all(source.spectrum.component2([1, 2, 3]*u.keV) == po2([1, 2, 3]*u.keV))

      one = source.spectrum.component1([1, 2, 3]*u.keV)
      two = source.spectrum.component2([1, 2, 3]*u.keV)

      #check spectral components
      assert np.all( np.abs(one + two - source.get_spatially_integrated_flux([1,2,3]*u.keV)) == 0 )
    
      #check spectral and spatial components
      #spatial = source.spatial_shape( ra*u.deg,dec*u.deg )
      spatial = source.spatial_shape( [ra, ra, ra]*u.deg, [dec, dec, dec]*u.deg )

      total = source( [ra, ra, ra]*u.deg, [dec, dec, dec]*u.deg, [1,2,3]*u.keV)
      spectrum = one + two
      assert np.all( np.abs( total - spectrum*spatial ) == 0 )
    
      total = source( [ra*1.01]*3*u.deg, [dec*1.01]*3*u.deg, [1,2,3]*u.keV)
      spectrum = one + two
      spatial = source.spatial_shape( [ra*1.01]*3*u.deg, [dec*1.01]*3*u.deg )
      assert np.all( np.abs( total - spectrum*spatial ) == 0 )
      
      model = Model( source )
      new_model = clone_model( model )
      
      new_total = new_model['test_source_%s' % name]( [ra*1.01]*3*u.deg, [dec*1.01]*3*u.deg, [1,2,3]*u.keV)
      assert np.all( np.abs( total - new_total ) == 0 )
     
    
    for key in _known_functions:

        if key in ["Latitude_galactic_diffuse" ]:
        #not testing latitude galactic diffuse for now.
            continue
        
        this_function = _known_functions[key]

        if this_function._n_dim == 2  and not this_function().is_prior:

            test_one(this_function, key)
            
    with pytest.raises(AssertionError):
        #this will fail because the Latitude_galactic_diffuse function isn't normalized.
        test_one(_known_functions["Latitude_galactic_diffuse"], "Latitude_galactic_diffuse")
   

  
 
def test_free_param():

    spectrum = Log_parabola()
    source = ExtendedSource("test_source", spatial_shape=Gaussian_on_sphere(), spectral_shape=spectrum)

    parameters = [spectrum.alpha, spectrum.beta, spectrum.piv, spectrum.K, source.spatial_shape.lat0, source.spatial_shape.lon0, source.spatial_shape.sigma]

    for param in parameters:
        param.free = False

    assert len(source.free_parameters) == 0

    for i, param in enumerate(parameters):
        param.free = True
        assert len(source.free_parameters) == i+1





import pkg_resources
import os


def _get_data_file_path(data_file):
    """
    Returns the absolute path to the required data files.

    :param data_file: relative path to the data file, relative to the astromodels/data path.
    So to get the path to data/dark_matter/gammamc_dif.dat you need to use data_file="dark_matter/gammamc_dif.dat"
    :return: absolute path of the data file
    """

    try:

        file_path = pkg_resources.resource_filename("astromodels", 'data/%s' % data_file)

    except KeyError:

        raise IOError("Could not read or find data file %s. Try reinstalling astromodels. If this does not fix your "
                      "problem, open an issue on github." % (data_file))

    else:

        return os.path.abspath(file_path)



from __future__ import division
from past.utils import old_div
import astropy.units as u

from astromodels.functions.function import Function3D, FunctionMeta

import numpy as np

from astromodels.utils.angular_distance import angular_distance_fast
from future.utils import with_metaclass


class Continuous_injection_diffusion_ellipse(with_metaclass(FunctionMeta, Function3D)):
    r"""
        description :

            Positron and electrons diffusing away from the accelerator

        latex : $\left(\frac{180^\circ}{\pi}\right)^2 \frac{1.2154}{\sqrt{\pi^3} r_{\rm diff} ({\rm angsep} ({\rm x, y, lon_0, lat_0})+0.06 r_{\rm diff} )} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm x, y, lon_0, lat_0})}{r_{\rm diff} ^2} \right)$

        parameters :

            lon0 :

                desc : Longitude of the center of the source
                initial value : 0.0
                min : 0.0
                max : 360.0

            lat0 :

                desc : Latitude of the center of the source
                initial value : 0.0
                min : -90.0
                max : 90.0

            rdiff0 :

                desc : Projected diffusion radius. The maximum allowed value is used to define the truncation radius.
                initial value : 1.0
                min : 0
                max : 20

            delta :

                desc : index for the diffusion coefficient
                initial value : 0.5
                min : 0.3
                max : 0.6
                fix : yes

            b :

                desc : b field strength in uG
                initial value : 3
                min : 1
                max : 10.
                fix : yes

            piv :

                desc : Pivot for the diffusion radius
                initial value : 2e10
                min : 0
                fix : yes

            piv2 :

                desc : Pivot for converting gamma energy to electron energy (always be 1 TeV)
                initial value : 1e9
                min : 0
                fix : yes

            incl :

                desc : inclination of semimajoraxis to a line of constant latitude
                initial value : 0.0
                min : -90.0
                max : 90.0
                fix : yes

            elongation :

                desc : elongation of the ellipse (b/a)
                initial value : 1.
                min : 0.1
                max : 10.

        """

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.rdiff0.unit = x_unit

        # Delta is of course unitless

        self.delta.unit = u.dimensionless_unscaled
        self.b.unit = u.dimensionless_unscaled
        self.incl.unit = x_unit
        self.elongation.unit = u.dimensionless_unscaled

        # Piv has the same unit as energy (which is z)

        self.piv.unit = z_unit
        self.piv2.unit = z_unit

    def evaluate(self, x, y, z, lon0, lat0, rdiff0, delta, b, piv, piv2, incl, elongation):

        lon, lat = x, y
        energy = z

        # energy in kev -> TeV.
        # NOTE: the use of piv2 is necessary to preserve dimensional correctness: the logarithm can only be taken
        # of a dimensionless quantity, so there must be a pivot there.

        e_energy_piv2 = 17. * np.power(old_div(energy, piv2), 0.54 + 0.046 * np.log10(old_div(energy, piv2)))
        e_piv_piv2 = 17. * np.power(old_div(piv, piv2), 0.54 + 0.046 * np.log10(old_div(piv, piv2)))

        try:

            rdiff_a = rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), old_div((delta - 1.), 2.)) * \
                    np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 * np.power(1. + 0.0107 * e_piv_piv2, -1.5)) / \
                    np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 * np.power(1. + 0.0107 * e_energy_piv2, -1.5))

        except ValueError:

            # This happens when using units, because astropy.units fails with the message:
            # "ValueError: Quantities and Units may only be raised to a scalar power"

            # Work around the problem with this loop, which is slow but using units is only for testing purposes or
            # single calls, so it shouldn't matter too much
            rdiff_a = np.array( [(rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), x)).value for x in (delta - 1.) / 2. * np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 * np.power(1. + 0.0107 * e_piv_piv2, -1.5)) /
                                  np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 * np.power(1. + 0.0107 * e_energy_piv2, -1.5))]) * rdiff0.unit

        rdiff_b = rdiff_a * elongation

        pi = np.pi

        angsep = angular_distance_fast(lon, lat, lon0, lat0)
        ang = np.arctan2(lat - lat0, (lon - lon0) * np.cos(lat0 * np.pi / 180.))

        theta = np.arctan2(old_div(np.sin(ang-incl*np.pi/180.),elongation), np.cos(ang-incl*np.pi/180.))

        rdiffs_a, thetas = np.meshgrid(rdiff_a, theta)
        rdiffs_b, angseps = np.meshgrid(rdiff_b, angsep)

        rdiffs = np.sqrt(rdiffs_a ** 2 * np.cos(thetas) ** 2 + rdiffs_b ** 2 * np.sin(thetas) ** 2)


        results = np.power(old_div(180.0, pi), 2) * 1.22 / (pi * np.sqrt(pi) * rdiffs_a * np.sqrt(elongation) * (angseps + 0.06 * rdiffs)) *  np.exp(old_div(-np.power(angseps, 2), rdiffs ** 2))

        return results


    def get_boundaries(self):

        # Truncate the function at the max of rdiff allowed

        maximum_rdiff = self.rdiff0.max_value

        min_latitude = max(-90., self.lat0.value - maximum_rdiff)
        max_latitude = min(90., self.lat0.value + maximum_rdiff)

        max_abs_lat = max(np.absolute(min_latitude), np.absolute(max_latitude))

        if max_abs_lat > 89. or old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.)) >= 180.:

            min_longitude = 0.
            max_longitude = 360.

        else:

            min_longitude = self.lon0.value - old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))
            max_longitude = self.lon0.value + old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))

            if min_longitude < 0.:

                min_longitude += 360.

            elif max_longitude > 360.:

                max_longitude -= 360.

        return (min_longitude, max_longitude), (min_latitude, max_latitude)

    def get_total_spatial_integral(self, z=None):  
        """
        Returns the total integral (for 2D functions) or the integral over the spatial components (for 3D functions).
        needs to be implemented in subclasses.

        :return: an array of values of the integral (same dimension as z).
        """

        if isinstance( z, u.Quantity):
            z = z.value
        return np.ones_like( z )


class Continuous_injection_diffusion(with_metaclass(FunctionMeta, Function3D)):
    r"""
        description :

            Positron and electrons diffusing away from the accelerator

        latex : $\left(\frac{180^\circ}{\pi}\right)^2 \frac{1.2154}{\sqrt{\pi^3} r_{\rm diff} ({\rm angsep} ({\rm x, y, lon_0, lat_0})+0.06 r_{\rm diff} )} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm x, y, lon_0, lat_0})}{r_{\rm diff} ^2} \right)$

        parameters :

            lon0 :

                desc : Longitude of the center of the source
                initial value : 0.0
                min : 0.0
                max : 360.0

            lat0 :

                desc : Latitude of the center of the source
                initial value : 0.0
                min : -90.0
                max : 90.0

            rdiff0 :

                desc : Projected diffusion radius limited by the cooling time. The maximum allowed value is used to define the truncation radius.
                initial value : 1.0
                min : 0
                max : 20

            rinj :

                desc : Ratio of diffusion radius limited by the injection time over rdiff0. The maximum allowed value is used to define the truncation radius.
                initial value : 100.0
                min : 0
                max : 200
                fix : yes

            delta :

                desc : index for the diffusion coefficient
                initial value : 0.5
                min : 0.3
                max : 0.6
                fix : yes

            b :

                desc : b field strength in uG
                initial value : 3
                min : 1
                max : 10.
                fix : yes

            piv :

                desc : Pivot for the diffusion radius
                initial value : 2e10
                min : 0
                fix : yes

            piv2 :
                desc : Pivot for converting gamma energy to electron energy (always be 1 TeV)
                initial value : 1e9
                min : 0
                fix : yes

        """

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.rdiff0.unit = x_unit
        self.rinj.unit = u.dimensionless_unscaled

        # Delta is of course unitless

        self.delta.unit = u.dimensionless_unscaled
        self.b.unit = u.dimensionless_unscaled

        # Piv has the same unit as energy (which is z)

        self.piv.unit = z_unit
        self.piv2.unit = z_unit

    def evaluate(self, x, y, z, lon0, lat0, rdiff0, rinj, delta, b, piv, piv2):

        lon, lat = x, y
        energy = z

        # energy in kev -> TeV.
        # NOTE: the use of piv2 is necessary to preserve dimensional correctness: the logarithm can only be taken
        # of a dimensionless quantity, so there must be a pivot there.

        e_energy_piv2 = 17. * np.power(old_div(energy, piv2), 0.54 + 0.046 * np.log10(old_div(energy, piv2)))
        e_piv_piv2 = 17. * np.power(old_div(piv, piv2), 0.54 + 0.046 * np.log10(old_div(piv, piv2)))

        rdiff_c = rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), old_div((delta - 1.), 2.)) * \
                np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 * np.power(1. + 0.0107 * e_piv_piv2, -1.5)) / \
                np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 * np.power(1. + 0.0107 * e_energy_piv2, -1.5))

        rdiff_i = rdiff0 * rinj * np.power(old_div(e_energy_piv2, e_piv_piv2), old_div(delta, 2.))

        rdiff = np.minimum(rdiff_c, rdiff_i)

        angsep = angular_distance_fast(lon, lat, lon0, lat0)

        pi = np.pi

        rdiffs, angseps = np.meshgrid(rdiff, angsep)

        return np.power(old_div(180.0, pi), 2) * 1.2154 / (pi * np.sqrt(pi) * rdiffs * (angseps + 0.06 * rdiffs)) * \
               np.exp(old_div(-np.power(angseps, 2), rdiffs ** 2))


    def get_boundaries(self):

        # Truncate the function at the max of rdiff allowed

        maximum_rdiff = self.rdiff0.max_value

        min_latitude = max(-90., self.lat0.value - maximum_rdiff)
        max_latitude = min(90., self.lat0.value + maximum_rdiff)

        max_abs_lat = max(np.absolute(min_latitude), np.absolute(max_latitude))

        if max_abs_lat > 89. or old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.)) >= 180.:

            min_longitude = 0.
            max_longitude = 360.

        else:

            min_longitude = self.lon0.value - old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))
            max_longitude = self.lon0.value + old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))

            if min_longitude < 0.:

                min_longitude += 360.

            elif max_longitude > 360.:

                max_longitude -= 360.

        return (min_longitude, max_longitude), (min_latitude, max_latitude)

    def get_total_spatial_integral(self, z=None):  
        """
        Returns the total integral (for 2D functions) or the integral over the spatial components (for 3D functions).
        needs to be implemented in subclasses.

        :return: an array of values of the integral (same dimension as z).
        """

        if isinstance( z, u.Quantity):
            z = z.value
        return np.ones_like( z )


class Continuous_injection_diffusion_legacy(with_metaclass(FunctionMeta, Function3D)):
    r"""
        description :

            Positron and electrons diffusing away from the accelerator

        latex : $\left(\frac{180^\circ}{\pi}\right)^2 \frac{1.2154}{\sqrt{\pi^3} r_{\rm diff} ({\rm angsep} ({\rm x, y, lon_0, lat_0})+0.06 r_{\rm diff} )} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm x, y, lon_0, lat_0})}{r_{\rm diff} ^2} \right)$

        parameters :

            lon0 :

                desc : Longitude of the center of the source
                initial value : 0.0
                min : 0.0
                max : 360.0

            lat0 :

                desc : Latitude of the center of the source
                initial value : 0.0
                min : -90.0
                max : 90.0

            rdiff0 :

                desc : Projected diffusion radius. The maximum allowed value is used to define the truncation radius.
                initial value : 1.0
                min : 0
                max : 20

            delta :

                desc : index for the diffusion coefficient
                initial value : 0.5
                min : 0.3
                max : 0.6
                fix : yes

            uratio :

                desc : ratio between u_cmb and u_B
                initial value : 0.5
                min : 0.01
                max : 100.
                fix : yes

            piv :

                desc : Pivot for the diffusion radius
                initial value : 2e10
                min : 0
                fix : yes

            piv2 :
                desc : Pivot for converting gamma energy to electron energy (always be 1 TeV)
                initial value : 1e9
                min : 0
                fix : yes

        """

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.rdiff0.unit = x_unit

        # Delta is of course unitless

        self.delta.unit = u.dimensionless_unscaled
        self.uratio.unit = u.dimensionless_unscaled

        # Piv has the same unit as energy (which is z)

        self.piv.unit = z_unit
        self.piv2.unit = z_unit

    def evaluate(self, x, y, z, lon0, lat0, rdiff0, delta, uratio, piv, piv2):

        lon, lat = x, y
        energy = z

        # energy in kev -> TeV.
        # NOTE: the use of piv2 is necessary to preserve dimensional correctness: the logarithm can only be taken
        # of a dimensionless quantity, so there must be a pivot there.

        e_energy_piv2 = 17. * np.power(old_div(energy, piv2), 0.54 + 0.046 * np.log10(old_div(energy, piv2)))
        e_piv_piv2 = 17. * np.power(old_div(piv, piv2), 0.54 + 0.046 * np.log10(old_div(piv, piv2)))

        try:

            rdiff = rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), old_div((delta - 1.), 2.)) * \
                    np.sqrt(1. + uratio * np.power(1. + 0.0107 * e_piv_piv2, -1.5)) / \
                    np.sqrt(1. + uratio * np.power(1. + 0.0107 * e_energy_piv2, -1.5))

        except ValueError:

            # This happens when using units, because astropy.units fails with the message:
            # "ValueError: Quantities and Units may only be raised to a scalar power"

            # Work around the problem with this loop, which is slow but using units is only for testing purposes or
            # single calls, so it shouldn't matter too much
            rdiff = np.array( [(rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), x)).value for x in (delta - 1.) / 2. * np.sqrt(1. + uratio * np.power(1. + 0.0107 * e_piv_piv2, -1.5)) /
                                  np.sqrt(1. + uratio * np.power(1. + 0.0107 * e_energy_piv2, -1.5))]) * rdiff0.unit

        angsep = angular_distance_fast(lon, lat, lon0, lat0)

        pi = np.pi

        rdiffs, angseps = np.meshgrid(rdiff, angsep)

        return np.power(old_div(180.0, pi), 2) * 1.2154 / (pi * np.sqrt(pi) * rdiffs * (angseps + 0.06 * rdiffs)) * \
               np.exp(old_div(-np.power(angseps, 2), rdiffs ** 2))


    def get_boundaries(self):

        # Truncate the function at the max of rdiff allowed

        maximum_rdiff = self.rdiff0.max_value

        min_latitude = max(-90., self.lat0.value - maximum_rdiff)
        max_latitude = min(90., self.lat0.value + maximum_rdiff)

        max_abs_lat = max(np.absolute(min_latitude), np.absolute(max_latitude))

        if max_abs_lat > 89. or old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.)) >= 180.:

            min_longitude = 0.
            max_longitude = 360.

        else:

            min_longitude = self.lon0.value - old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))
            max_longitude = self.lon0.value + old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))

            if min_longitude < 0.:

                min_longitude += 360.

            elif max_longitude > 360.:

                max_longitude -= 360.

        return (min_longitude, max_longitude), (min_latitude, max_latitude)

    def get_total_spatial_integral(self, z=None):  
        """
        Returns the total integral (for 2D functions) or the integral over the spatial components (for 3D functions).
        needs to be implemented in subclasses.

        :return: an array of values of the integral (same dimension as z).
        """

        if isinstance( z, u.Quantity):
            z = z.value
        return np.ones_like( z )




import numpy as np

def vincenty(lon0, lat0, a1, s):
  """
  Returns the coordinates of a new point that is a given angular distance s away from a starting point (lon0, lat0) at bearing (angle from north) a1), to within a given precision

  Note that this calculation is a simplified version of the full vincenty problem, which solves for the coordinates on the surface on an arbitrary ellipsoid. Here we only care about the surface of a sphere.

  Note: All parameters are assumed to be given in DEGREES
  :param lon0: float, longitude of starting point
  :param lat0: float, latitude of starting point
  :param a1: float, bearing to second point, i.e. angle between due north and line connecting 2 points
  :param s: float, angular distance between the two points
  :return: coordinates of second point in degrees
  """

  lon0 = np.deg2rad(lon0)
  lat0 = np.deg2rad(lat0)
  a1   = np.deg2rad(a1)
  s    = np.deg2rad(s)

  sina = np.cos(lat0) * np.sin(a1)

  num1 = np.sin(lat0)*np.cos(s) + np.cos(lat0)*np.sin(s)*np.cos(a1)
  den1 = np.sqrt(sina**2 + (np.sin(lat0)*np.sin(s) - np.cos(lat0)*np.cos(a1))**2)
  lat  = np.rad2deg(np.arctan2(num1, den1))

  num2 = np.sin(s)*np.sin(a1)
  den2 = np.cos(lat0)*np.cos(s) - np.sin(lat0)*np.sin(s)*np.cos(a1)
  L    = np.arctan2(num2, den2)
  lon  = np.rad2deg(lon0 + L)

  return lon, lat



from functools import wraps


try:
    from functools import lru_cache
except ImportError:

    def lru_cache(user_function, **kwargs):
        cache = {}

        @wraps(user_function)
        def wrapper(*args):
            key = tuple(args)
            if key not in cache:
                cache[key] = user_function(*args)
            return cache[key]

        return wrapper


import astropy.units as astropy_units
import numpy as np

import sys





is_py3 = True

if sys.version_info[0] < 3:

    is_py3 = False

def cache_array_method(*args, **kwargs):
    """
    LRU cache implementation for methods whose FIRST parameter is a numpy array
    modified from: https://gist.github.com/Susensio/61f4fee01150caaac1e10fc5f005eb75
    """

    def decorator(function):
        @wraps(function)
        def wrapper(s, np_array, *args, **kwargs):
            hashable_array = tuple(np_array)
            return cached_wrapper(s, hashable_array, *args, **kwargs)

        @lru_cache(*args, **kwargs)
        def cached_wrapper(s, hashable_array, *args, **kwargs):
            array = np.array(hashable_array)
            return function(s, array, *args, **kwargs)

        # copy lru_cache attributes over too
        wrapper.cache_info = cached_wrapper.cache_info
        wrapper.cache_clear = cached_wrapper.cache_clear
        return wrapper

    return decorator


try:

    import pyatomdb

    has_atomdb = True

except:

    has_atomdb = False


from astromodels.functions.function import Function1D, FunctionMeta
from astromodels.utils import configuration
from astromodels.utils.data_files import _get_data_file_path

from astropy.io import fits
import os
import six


if has_atomdb:
    # APEC class
    @six.add_metaclass(FunctionMeta)
    class APEC(Function1D):
        r"""
        description :
            The Astrophysical Plasma Emission Code (APEC, Smith et al. 2001)
            contributed by Dominique Eckert
        parameters :
            K :
                desc : Normalization in units of 1e-14/(4*pi*(1+z)^2*dA*2)*EM
                initial value : 1.0
                is_normalization : True
                transformation : log10
                min : 1e-30
                max : 1e3
                delta : 0.1
            kT :
                desc : Plasma temperature
                initial value : 1.0
                min : 0.08
                max : 64
                delta : 0.1
            abund :
                desc : Metal abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            redshift :
                desc : Source redshift
                initial value : 0.1
                min : 0.0
                max : 10.0
                delta : 1e-3
                fix : yes

        """

        def _set_units(self, x_unit, y_unit):
            self.kT.unit = astropy_units.keV

            self.abund.unit = astropy_units.dimensionless_unscaled

            self.redshift.unit = astropy_units.dimensionless_unscaled

            self.K.unit = y_unit

        def init_session(self, abund_table="AG89"):
            # initialize PyAtomDB session
            self.session = pyatomdb.spectrum.CIESession(abundset=abund_table)

        def evaluate(self, x, K, kT, abund, redshift):
            assert self.session is not None, "please run init_session(abund)"

            sess = self.session

            nval = len(x)

            xz = x * (1.0 + redshift)

            ebplus = (np.roll(xz, -1) + xz)[: nval - 1] / 2.0

            ebounds = np.empty(nval + 1)

            ebounds[1:nval] = ebplus

            ebounds[0] = xz[0] - (ebplus[0] - xz[0])

            ebounds[nval] = xz[nval - 1] + (xz[nval - 1] - ebplus[nval - 2])

            binsize = (np.roll(ebounds, -1) - ebounds)[:nval]

            sess.set_response(ebounds, raw=True)

            sess.set_abund(
                [
                    6,
                    7,
                    8,
                    9,
                    10,
                    11,
                    12,
                    13,
                    14,
                    16,
                    17,
                    18,
                    19,
                    20,
                    21,
                    22,
                    23,
                    24,
                    25,
                    26,
                    27,
                    28,
                    29,
                    30,
                ],
                abund,
            )

            spec = sess.return_spectrum(kT) / binsize / 1e-14

            return K * spec

    # VAPEC class
    @six.add_metaclass(FunctionMeta)
    class VAPEC(Function1D):
        r"""
        description :
            The Astrophysical Plasma Emission Code (APEC, Smith et al. 2001), variable abundances for individual elements
            contributed by Dominique Eckert
        parameters :
            K :
                desc : Normalization in units of 1e-14/(4*pi*(1+z)^2*dA*2)*EM
                initial value : 1.0
                is_normalization : True
                transformation : log10
                min : 1e-30
                max : 1e3
                delta : 0.1
            kT :
                desc : Plasma temperature
                initial value : 1.0
                min : 0.08
                max : 64
                delta : 0.1
            Fe :
                desc : Fe abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            C :
                desc : C abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            N :
                desc : N abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            O :
                desc : O abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            Ne :
                desc : Ne abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            Mg :
                desc : Mg abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            Al :
                desc : Al abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            Si :
                desc : Si abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            S :
                desc : S abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            Ar :
                desc : Ar abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            Ca :
                desc : Ca abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            Ni :
                desc : Ni abundance
                initial value : 1
                min : 0.0
                max : 5.0
                delta : 0.01
                fix : yes
            redshift :
                desc : Source redshift
                initial value : 0.1
                min : 0.0
                max : 10.0
                delta : 1e-3
                fix : yes

        """

        def _set_units(self, x_unit, y_unit):
            self.kT.unit = astropy_units.keV

            self.Fe.unit = astropy_units.dimensionless_unscaled

            self.C.unit = astropy_units.dimensionless_unscaled

            self.N.unit = astropy_units.dimensionless_unscaled

            self.O.unit = astropy_units.dimensionless_unscaled

            self.Ne.unit = astropy_units.dimensionless_unscaled

            self.Mg.unit = astropy_units.dimensionless_unscaled

            self.Al.unit = astropy_units.dimensionless_unscaled

            self.Si.unit = astropy_units.dimensionless_unscaled

            self.Ar.unit = astropy_units.dimensionless_unscaled

            self.Ca.unit = astropy_units.dimensionless_unscaled

            self.Ni.unit = astropy_units.dimensionless_unscaled

            self.redshift.unit = astropy_units.dimensionless_unscaled

            self.K.unit = y_unit

        def init_session(self, abund_table="AG89"):
            # initialize PyAtomDB session
            self.session = pyatomdb.spectrum.CIESession(abundset=abund_table)

        def evaluate(
            self, x, K, kT, Fe, C, N, O, Ne, Mg, Al, Si, S, Ar, Ca, Ni, redshift
        ):
            assert self.session is not None, "please run init_session(abund)"

            sess = self.session

            nval = len(x)

            xz = x * (1.0 + redshift)

            ebplus = (np.roll(xz, -1) + xz)[: nval - 1] / 2.0

            ebounds = np.empty(nval + 1)

            ebounds[1:nval] = ebplus

            ebounds[0] = xz[0] - (ebplus[0] - xz[0])

            ebounds[nval] = xz[nval - 1] + (xz[nval - 1] - ebplus[nval - 2])

            binsize = (np.roll(ebounds, -1) - ebounds)[:nval]

            sess.set_response(ebounds, raw=True)

            sess.set_abund(
                [6,], C,
            )

            sess.set_abund(
                [7,], N,
            )

            sess.set_abund(
                [8,], O,
            )

            sess.set_abund(
                [10,], Ne,
            )

            sess.set_abund(
                [12,], Mg,
            )

            sess.set_abund(
                [13,], Al,
            )

            sess.set_abund(
                [14,], Si,
            )

            sess.set_abund(
                [16,], S,
            )

            sess.set_abund(
                [18,], Ar,
            )

            sess.set_abund(
                [20,], Ca,
            )

            sess.set_abund(
                [26,], Fe,
            )

            sess.set_abund(
                [28,], Ni,
            )

            sess.set_abund(
                [9, 11, 15, 17, 19, 21, 22, 23, 24, 25, 27, 29, 30], Fe
            )  # Remaining elements are set to Fe

            spec = sess.return_spectrum(kT) / binsize / 1e-14

            return K * spec



if is_py3:


    _abs_tables = {
        "phabs": {"AG89": "angr", "ASPL": "aspl"},
        "tbabs": {"AG89": "angr", "ASPL": "aspl", "WILM": "wilm"},
        "wabs": {"AG89": "angr"},
    }
    _abund_info = {}
    _abund_info[
        "WILMS"
    ] = "wilms\nfrom Wilms, Allen & McCray (2000), ApJ 542, 914 \n except for elements not listed which are given zero abundance)\n https://heasarc.nasa.gov/xanadu/xspec/manual/XSabund.html "
    _abund_info[
        "AG89"
    ] = "angr\nfrom Anders E. & Grevesse N. (1989, Geochimica et Cosmochimica Acta 53, 197)\n https://heasarc.nasa.gov/xanadu/xspec/manual/XSabund.html"
    _abund_info[
        "ASPL"
    ] = "aspl\nfrom Asplund M., Grevesse N., Sauval A.J. & Scott P. (2009, ARAA, 47, 481)\nhttps://heasarc.nasa.gov/xanadu/xspec/manual/XSabund.html"


    def _get_xsect_table(model, abund_table):
        """
        contructs the abundance table from the values given
        """

        assert model in _abs_tables, "the model %s does not exist" % model
        assert abund_table in _abs_tables[model], (
            "the table %s does not exist" % abund_table
        )

        path_to_xsect = _get_data_file_path(
            os.path.join(
                "xsect", "xsect_%s_%s.fits" % (model, _abs_tables[model][abund_table])
            )
        )

        fxs = fits.open(path_to_xsect)
        dxs = fxs[1].data
        xsect_ene = dxs["ENERGY"]
        xsect_val = dxs["SIGMA"]

        return np.array(xsect_ene), np.array(xsect_val)




    # PhAbs class
    @six.add_metaclass(FunctionMeta)
    class PhAbs(Function1D):
        r"""
        description :
            Photometric absorption (phabs implementation), f(E) = exp(- NH * sigma(E))
            contributed by Dominique Eckert
        parameters :
            NH :
                desc : absorbing column density in units of 1e22 particles per cm^2
                initial value : 1.0
                is_normalization : False
                transformation : log10
                min : 1e-4
                max : 1e4
                delta : 0.1

            redshift :
                desc : the redshift of the source
                initial value : 0.
                is_normalization : False
                min : 0
                max : 15
                delta : 0.1
                fixed: True


        """

        def _setup(self):
            self._fixed_units = (astropy_units.keV, astropy_units.dimensionless_unscaled)
            self.init_xsect()

        def _set_units(self, x_unit, y_unit):
            self.NH.unit = astropy_units.cm ** (-2)
            self.redshift.unit = astropy_units.dimensionless_unscaled

        def init_xsect(self, abund_table="AG89"):
            """
            Set the abundance table

            :param abund_table: "ASPL", "AG89" 
            :returns: 
            :rtype: 

            """

            # load cross section data

            try:
                self.xsect_ene, self.xsect_val = _get_xsect_table("phabs", abund_table)
                self._abund_table = abund_table

            except:

                print("defaulting to AG89")
                self.xsect_ene, self.xsect_val = _get_xsect_table("phabs", abund_table)

                self._abund_table = "AG89"

        @cache_array_method()
        def _cached_interp(self, x):

            return np.interp(x, self.xsect_ene, self.xsect_val)

        def evaluate(self, x, NH, redshift):

            if isinstance(x, astropy_units.Quantity):

                _unit = astropy_units.cm ** 2
                _y_unit = astropy_units.dimensionless_unscaled
                _x = x.value

            else:

                _unit = 1.0
                _y_unit = 1.0

                _x = x

            xsect_interp = self._cached_interp(_x * (1 + redshift))

            spec = np.exp(-NH * xsect_interp * _unit) * _y_unit

            return spec


    # TbAbs class
    @six.add_metaclass(FunctionMeta)
    class TbAbs(Function1D):
        r"""
        description :
            Photometric absorption (Tbabs implementation), f(E) = exp(- NH * sigma(E))
            contributed by Dominique Eckert
        parameters :
            NH :
                desc : absorbing column density in units of 1e22 particles per cm^2
                initial value : 1.0
                is_normalization : True
                transformation : log10
                min : 1e-4
                max : 1e4
                delta : 0.1

            redshift :
                desc : the redshift of the source
                initial value : 0.
                is_normalization : False
                min : 0
                max : 15
                delta : 0.1
                fixed: True


        """

        def _setup(self):

            self.init_xsect()

            self._fixed_units = (astropy_units.keV, astropy_units.dimensionless_unscaled)

        def _set_units(self, x_unit, y_unit):
            self.NH.unit = astropy_units.cm ** (-2)
            self.redshift.unit = astropy_units.dimensionless_unscaled

        def init_xsect(self, abund_table="WILM"):
            """
            Set the abundance table

            :param abund_table: "WILM", "ASPL", "AG89" 
            :returns: 
            :rtype: 

            """

            try:
                self.xsect_ene, self.xsect_val = _get_xsect_table("tbabs", abund_table)
                self._abund_table = abund_table

            except:

                print("defaulting to WILM")
                self.xsect_ene, self.xsect_val = _get_xsect_table("tbabs", abund_table)

                self._abund_table = "WILM"

        @property
        def abundance_table(self):
            print(_abund_info[self._abund_table])

        @cache_array_method()
        def _cached_interp(self, x):

            return np.interp(x, self.xsect_ene, self.xsect_val)

        def evaluate(self, x, NH, redshift):

            if isinstance(x, astropy_units.Quantity):

                _unit = astropy_units.cm ** 2
                _y_unit = astropy_units.dimensionless_unscaled
                _x = x.value

            else:

                _unit = 1.0
                _y_unit = 1.0

                _x = x

            xsect_interp = self._cached_interp(_x * (1 + redshift))

            spec = np.exp(-NH * xsect_interp * _unit) * _y_unit

            return spec


    # WAbs class
    @six.add_metaclass(FunctionMeta)
    class WAbs(Function1D):
        r"""
        description :
            Photometric absorption (Wabs implementation), f(E) = exp(- NH * sigma(E))
            contributed by Dominique Eckert
        parameters :
            NH :
                desc : absorbing column density in units of 1e22 particles per cm^2
                initial value : 1.0
                is_normalization : True
                transformation : log10
                min : 1e-4
                max : 1e4
                delta : 0.1
            redshift :
                desc : the redshift of the source
                initial value : 0.
                is_normalization : False
                min : 0
                max : 15
                delta : 0.1
                fixed: True


        """

        def _setup(self):
            self._fixed_units = (astropy_units.keV, astropy_units.dimensionless_unscaled)
            self.init_xsect()

        def _set_units(self, x_unit, y_unit):
            self.NH.unit = astropy_units.cm ** (-2)
            self.redshift.unit = astropy_units.dimensionless_unscaled

        def init_xsect(self):
            """
            Set the abundance table

            :returns:
            :rtype:

            """

            self.xsect_ene, self.xsect_val = _get_xsect_table("wabs", "AG89")

            self._abund_table = "AG89"

        @property
        def abundance_table(self):
            print(_abund_info[self._abund_table])

        @cache_array_method()
        def _cached_interp(self, x):

            return np.interp(x, self.xsect_ene, self.xsect_val)

        def evaluate(self, x, NH, redshift):

            if isinstance(x, astropy_units.Quantity):

                _unit = astropy_units.cm ** 2
                _y_unit = astropy_units.dimensionless_unscaled
                _x = x.value

            else:

                _unit = 1.0
                _y_unit = 1.0

                _x = x

            xsect_interp = self._cached_interp(_x * (1 + redshift))

            spec = np.exp(-NH * xsect_interp * _unit) * _y_unit

            return spec



from __future__ import print_function
import collections
import sys
import time
import astropy.units as u
import os
import re
import warnings

from astromodels.core.my_yaml import my_yaml
from astromodels.functions.function import get_function_class
from astromodels.utils.configuration import get_user_data_path


class XSpecNotAvailable(ImportWarning):
    pass

from astromodels.xspec import _xspec

# When running in a Anaconda environment, the package xspec-modelsonly
# will install the models data in a specific place, so we set the HEADAS variable to point to the right place for
# the models to find their data
if os.environ.get("CONDA_PREFIX") is not None and os.environ.get("HEADAS", None) is None:

    # NOTE: the 'headas' directory doesn't actually exists, but models are looked for (according to Xspec documentation)
    # into $HEADAS../spectral, so we put HEADAS=[....]/headas so that $HEADAS/../ will be the right place where
    # the 'spectral' directory is

    os.environ['HEADAS'] = os.path.join(os.environ.get("CONDA_PREFIX"), 'lib', 'Xspec', 'headas')

    #os.environ['HEADAS'] = os.path.join(os.environ.get("CONDA_PREFIX"), 'Xspec', 'headas')

    # we need to create the directory otherwise an exception casted:
    if not os.path.exists(os.environ['HEADAS']):
        os.makedirs(os.environ['HEADAS'])

# This list defines all python protected names (names which variables or attributes should not have)
illegal_variable_names_ = 'and, assert, break, class, continue, def, del, elif, else, except, exec, finally, for,' \
                         'if, import, in, is, lambda, not, or, pass, print, raise, return, try, while, data, ' \
                         'float, int, numeric, array, close, input, open, range, type, write, zeros, from, global'
# Make it a list
illegal_variable_names = illegal_variable_names_.replace(" ","").split(",")


def find_model_dat():
    """
    Find the file containing the definition of all the models in Xspec
    (model.dat) and return its path
    """

    # model.dat is in $HEADAS/../spectral

    headas_env = os.environ.get("HEADAS")

    assert headas_env is not None, ("You need to setup the HEADAS variable before importing this module."
                                    " See Heasoft documentation.")

    # Expand all variables and other things like ~
    headas_env = os.path.expandvars(os.path.expanduser(headas_env))

    # Lazy check that it exists

    assert os.path.exists(headas_env), "The HEADAS env. variable point to a non-existent directory: %s" % (headas_env)

    # Get one directory above HEADAS (i.e., $HEADAS/..)

    inferred_path = os.path.dirname(headas_env)

    # Now model.dat should be in $HEADAS/../spectral/manager

    final_path = os.path.join(inferred_path, 'spectral', 'manager', 'model.dat')

    # Check that model.dat exists

    assert os.path.exists(final_path), "Cannot find Xspec model definition file %s" % (final_path)

    return os.path.abspath(final_path)


def get_models(model_dat_path):
    """
    Parse the model.dat file from Xspec and returns a dictionary containing the definition of all the models

    :param model_dat_path: the path to the model.dat file
    :return: dictionary containing the definition of all XSpec models
    """

    # Check first if we already have a model data file in the data directory


    with open(model_dat_path) as f:

        # model.dat is a text file, no size issues here (will fit in memory)

        model_dat = f.read()

    # Replace \r or \r\n with \n (the former two might be in the file if
    # it originates from Windows)
    if "\n" in model_dat:

        model_dat = model_dat.replace("\r", "")

    else:

        model_dat = model_dat.replace("\r", "\n")

    # Loop through the file and build the model definition
    # dictionary

    lines = model_dat.split("\n")

    model_definitions = collections.OrderedDict()

    for line in lines:

        match = re.match('''(.+(add|mul|con|acn).+)''', line)

        if match is not None:

            # This is a model definition

            tokens = line.split()

            if len(tokens) == 7:

                (model_name, n_parameters,
                 min_energy, max_energy,
                 library_function,
                 model_type,
                 flag) = line.split()

            else:

                (model_name, n_parameters,
                 min_energy, max_energy,
                 library_function,
                 model_type,
                 flag,
                 flag_2) = line.split()

            this_model = collections.OrderedDict()

            this_model['description'] = 'The %s model from XSpec (https://heasarc.gsfc.nasa.gov/xanadu/' \
                                                   'xspec/manual/XspecModels.html)' % model_name

            this_model['parameters'] = collections.OrderedDict()

            model_definitions[(model_name, library_function, model_type)] = this_model

        else:

            # This is a parameter definition

            if len(line.split()) == 0:
                # Empty line
                continue

            # Parameters are free by default, unless they have delta < 0
            # they are switch parameters or scale parameters

            free = True

            if line[0] == '$':

                # Probably a switch parameter

                free = False

                tokens = line.split()

                if len(tokens) == 2:

                    par_name = tokens[0][1:]

                    default_value = tokens[1]

                    par_unit = ""

                    hard_minimum, soft_minimum, soft_maximum, hard_maximum = (0, 0, 1e9, 1e9)

                elif len(tokens) == 3:

                    par_name = tokens[0][1:]

                    default_value = tokens[2]

                    par_unit = ""

                    hard_minimum, soft_minimum, soft_maximum, hard_maximum = (0, 0, 1e9, 1e9)

                else:

                    match = re.match('(\S+)\s+(\".+\"|[a-zA-Z]+)?(.+)*', line[1:])

                    par_name, par_unit, par_spec = match.groups()

                    tokens = par_spec.split()

                    if len(tokens) == 1:

                        default_value = tokens[0]
                        par_unit = ""
                        hard_minimum, soft_minimum, soft_maximum, hard_maximum = (0, 0, 1e9, 1e9)

                    else:

                        par_unit = ""

                        (default_value,
                         hard_minimum, soft_minimum,
                         soft_maximum, hard_maximum,
                         delta) = par_spec.split()


            else:

                # A normal parameter

                match = re.match('(\S+)\s+(\".+\"|\S+)(.+)', line)

                if match is None:

                    raise RuntimeError("Cannot parse parameter %s" % line)

                par_name, par_unit, par_spec = match.groups()

                if par_name[0] == '*':

                    # Scale parameter (always frozen)

                    par_name = par_name[1:]

                    free = False

                    tokens = par_spec.split()

                    if len(tokens) == 1:

                        default_value = tokens[0]
                        (hard_minimum, soft_minimum,
                         soft_maximum, hard_maximum,
                         delta) = (None, None, None, None, 0.1)

                    else:

                        (default_value,
                         hard_minimum, soft_minimum,
                         soft_maximum, hard_maximum,
                         delta) = par_spec.split()

                        delta = abs(float(delta))

                else:

                    (default_value,
                     hard_minimum, soft_minimum,
                     soft_maximum, hard_maximum,
                     delta) = par_spec.split()

                    delta = float(delta)

                    if delta <= 0:
                        free = False

                        delta = abs(delta)

            # Now fix the parameter name removing illegal characters
            # (for example 'log(A)' is not a legal name)

            par_name = re.sub('[\(,\)]', '_', par_name)
            par_name = re.sub('<', '_less_', par_name)
            par_name = re.sub('>', '_more_', par_name)
            par_name = re.sub('/', '_div_', par_name)
            par_name = re.sub('\-', '_minus_', par_name)
            par_name = re.sub('\+', '_plus_', par_name)
            par_name = re.sub('\.', '_dot_', par_name)
            par_name = re.sub('@', '_at_', par_name)

            # Parameter names must be lower case
            par_name = par_name.lower()

            # Some parameters are enclosed between ", like "z"
            par_name = par_name.replace('"', '')

            # "z" is a protected name in astromodels.
            if par_name == "z":

                par_name = "redshift"

                # this is illegal because of the 2D functions
            if par_name == 'y':

                par_name = 'y1'

            # Check that the parameter name is not an illegal Python name
            if par_name in illegal_variable_names:

                par_name = "xs_%s" % par_name

            # Sometimes the unit is " " which is not recognized by astropy
            if par_unit:

                par_unit = par_unit.replace("\"",'')

            # There are some errors in model.dat , like KeV instead of keV, and ergcm/s instead of erg cm /s
            # Let's correct them
            if par_unit == "KeV":

                par_unit = "keV"

            elif par_unit == "ergcm/s":

                par_unit = "erg cm / s"

            elif par_unit == "days":

                par_unit = "day"

            elif par_unit == "z" and par_name.lower() == "redshift":

                par_unit = ""



            # There are funny units in model.dat, like "Rs" (which means Schwarzschild radius) or other things
            # so let's try to convert the par_unit into an astropy.Unit instance. If that fails, use a unitless unit
            try:

                _ = u.Unit(par_unit)

            except ValueError:

                # Unit not recognized

                #warnings.warn("Unit %s is not recognized by astropy." % par_unit)

                par_unit = ''

            # Make sure that this is a valid python identifier
            # by matching it with the relative regexp
            if re.match('([a-zA-Z_][a-zA-Z0-9_]*)$', par_name) is None:

                raise ValueError("Illegal identifier name %s" % (par_name))


            if hard_maximum is not None and hard_minimum is not None:

                if (float(hard_maximum) < float(hard_minimum)):

                    raise ValueError("Hard maximum (%s) < hard minimum (%s)" %(hard_maximum,hard_minimum))

                if float(default_value) > float(hard_maximum):

                    default_value = hard_maximum

                elif float(default_value) < float(hard_minimum):

                    default_value = hard_minimum



            this_model['parameters'][par_name] = {'initial value': float(default_value),
                                                  'desc': '(see https://heasarc.gsfc.nasa.gov/xanadu/xspec/manual/'
                                                          'XspecModels.html)',
                                                  'min': hard_minimum,
                                                  'max': hard_maximum,
                                                  'delta': float(delta),
                                                  'unit': par_unit,
                                                  'free': free}

    return model_definitions

# TODO: gabs is returning all 1.0

class_definition_code = '''

from astromodels.functions.function import FunctionMeta, Function1D
import numpy as np
import astropy.units as u
from astromodels.xspec import _xspec
import six

# These are multiplicative functions which need numerical differentiation
_force_differentiation = ['XS_gabs', 'XS_expfac', 'XS_plabs', 'XS_pwab',
                          'XS_spline', 'XS_swind1', 'XS_xion', 'XS_zxipcf',
                          'XS_cabs', 'XS_wabs', 'XS_zwabs'
                          ]

@six.add_metaclass(FunctionMeta)
class XS_$MODEL_NAME$(Function1D):

    """
$DOCSTRING$
    """

    def _setup(self):

        # Link to the Xspec function
        self._model = _xspec.$XSPEC_FUNCTION$
        self._model_type = '$MODEL_TYPE$'
        
        # Decide whether we need numerical differentiation
        
        if self._model_type == 'add' or self._name in _force_differentiation:

            self._scale = 1e5
            
            self._differentiate = True

        else:

            # For multiplicative models, usually there is no differentiation to be made
            # (but there are exceptions, such as the gabs model)

            self._scale = 1
            
            self._differentiate = False
        
        # Now set the units as appropriate
        
        if self._model_type == 'add':
            
            self._fixed_units = (u.keV, 1 / (u.keV * u.cm**2 * u.s))
        
        else:
            
            self._fixed_units = (u.keV, u.dimensionless_unscaled)

    def evaluate(self, x, $PARAMETERS_NAMES$):

        quantity = False

        if isinstance(x, u.Quantity):

            x = np.array(x.to('keV').value, ndmin=1, copy=False, dtype=float)

            quantity = True

            parameters_tuple = tuple(map(lambda x:x.value, ($PARAMETERS_NAMES$,)))

        else:

            # Create a tuple of the current values of the parameters

            parameters_tuple = ($PARAMETERS_NAMES$,)
        
        # We need to make sure that the energy array is sorted because otherwise some Xspec models will give
        # incorrect values
        idx = np.argsort(x)
        
        # This is needed to be able to "reverse" the sort operation
        rev_idx = np.argsort(idx)
        
        # Ordered input vector
        
        xx = x[idx]
        
        if self._differentiate:

            # Finite difference differentiation of the Xspec
            # function. Indeed, the xspec function return the integral
            # of the function on energy ranges, while we need the
            # differential flux.

            # Adapt the epsilon to the value to reduce the error

            epsilon = xx / self._scale

            # In the normal case, when x is an array of more than one
            # element, the first call will succeed. If however there
            # is only one element in the array xspec will complain,
            # so handle that as a special case

            try:

                val = self._model(parameters_tuple, xx - epsilon, xx + epsilon)

            except TypeError:

                assert xx.shape[0]==1, "This is a bug, xspec call failed and x is not only one element"

                val = np.array(self._model(parameters_tuple, ((xx - epsilon)[0], (xx + epsilon)[0]))[0], ndmin=1)
            
            if self._model_type == 'add':
            
                # val is now F(x-epsilon,x+epsilon) ~ f(x) * ( 2 * epsilon )
                
                # In a additive model the function returns the integral over the bins
    
                final_value = val / (2 * epsilon)
            
            else:
            
                final_value = val

        else:

            # In a multiplicative model usually the function returns the average factor over
            # the bins (but there are exceptions, handled through the list _force_differentiation)

            try:
            
                final_value = self._model(parameters_tuple, xx, xx)

            except TypeError:

                assert xx.shape[0]==1, "This is a bug, xspec call failed and x is not only one element"

                final_value = self._model(parameters_tuple, ((xx)[0], (xx)[0]))


        if quantity:

            if self._model_type == 'add':

                return final_value[rev_idx] * u.Unit('1 / (keV cm^2 s)')

            else:

                return final_value[rev_idx] * u.dimensionless_unscaled

        else:

            return final_value[rev_idx]

    def has_fixed_units(self):

        return True

    def _set_units(self, x_unit, y_unit):

        # Make sure this is an energy
        assert str(x_unit.physical_type) == 'energy', "Trying to set x-unit with (%s), which is not energy" % (x_unit)

        # Make sure the y_unit is the correct one
        try:

            y_unit.in_units(self._fixed_units[1])

        except:

            raise RuntimeError("Xspec model %s cannot have units of %s" % (self.name, y_unit))

    def _integral(self, low_bounds, hi_bounds, $PARAMETERS_NAMES$):

        # Create a tuple of the current values of the parameters
        parameters_tuple = ($PARAMETERS_NAMES$,)

        return self._model(parameters_tuple, low_bounds, hi_bounds)

'''


def xspec_model_factory(model_name, xspec_function, model_type, definition):

    class_name = 'XS_%s' % model_name

    # Get the path to the user data directory
    user_data_path = get_user_data_path()

    # Check if the code for this function already exists

    code_file_name = os.path.join(user_data_path, '%s.py' % class_name)

    if os.path.exists(code_file_name):

        # Code already exists
        pass

    else:

        print("Generating code for Xspec model %s..." % model_name)

        # If this is an additive model (model_type == 'add') we need to add
        # one more parameter (normalization)

        if model_type == 'add':

            definition['parameters']['norm'] = {'initial value': 1.0,
                                                      'desc': '(see https://heasarc.gsfc.nasa.gov/xanadu/xspec/manual/'
                                                              'XspecModels.html)',
                                                      'min': 0,
                                                      'max': None,
                                                      'delta': 0.1,
                                                      'unit': 'keV / (cm2 s)',
                                                      'free': True}

        assert model_type != 'con', "Convolution models are not yet supported"

        # Get a list of the parameter names
        parameters_names = ", ".join(list(definition['parameters'].keys()))

        # Create the docstring
        docstring = my_yaml.dump(definition, default_flow_style=False)

        # Create the class by substituting in the class_definition_code the
        # relevant things for this model

        code = class_definition_code.replace('$MODEL_NAME$', model_name)
        code = code.replace('$DOCSTRING$', docstring)
        code = code.replace('$PARAMETERS_NAMES$', parameters_names)
        code = code.replace('$XSPEC_FUNCTION$', xspec_function)
        code = code.replace('$MODEL_TYPE$', model_type)

        # Write to the file

        with open(code_file_name, 'w+') as f:

            f.write("# This code has been automatically generated. Do not edit.\n")
            f.write("\n\n%s\n" % code)
        
        time.sleep(1)

    # Add the path to sys.path if it doesn't
    if user_data_path not in sys.path:

        sys.path.append(user_data_path)

    # Import the class in the current namespace (locals)
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        exec('from %s import %s' % (class_name, class_name))

    # Return the class we just created
    return class_name, locals()[class_name]

def setup_xspec_models():

    classes = []

    sys.stdout.write("Loading xspec models...")

    all_models = get_models(find_model_dat())

    for (model_name, xspec_function, model_type) in all_models:

        if model_type == 'con':

            # convolution models are not supported
            continue


        if not hasattr(_xspec, xspec_function):

            # Some function do not exist in the wrapper. Let's ignore them

            continue

        this_model = all_models[(model_name, xspec_function, model_type)]

        # When the class is created it is registered among the known functions in the function module
        # (it happens in the metaclass), so we don't need to do anything special here after the
        # class type is created

        this_class_name, this_class = xspec_model_factory(model_name, xspec_function, model_type, this_model)

        classes.append(this_class_name)

    sys.stdout.write("done\n")

    return classes

# This will either work or issue a warning if XSpec is not available

new_functions = setup_xspec_models()

# Now import the new classes in the local namespace (if any)
# This is needed to make the classes pickeable

__all__ = []

for function_name in new_functions:

    __all__.append(function_name)

    locals()[function_name] = get_function_class(function_name)



from builtins import str
from builtins import object
__author__ = 'giacomov'

import collections

PARTICLE_SOURCE = 'particle source'
POINT_SOURCE = 'point source'
EXTENDED_SOURCE = 'extended source'


class UnknownSourceType(Exception):
    pass


class Source(object):

    def __init__(self, list_of_components, src_type, spatial_shape=None):

        # Make the dictionary of components
        self._components = collections.OrderedDict()

        for component in list_of_components:

            self._components[component.name] = component

        if src_type not in (PARTICLE_SOURCE, POINT_SOURCE, EXTENDED_SOURCE):

            raise UnknownSourceType("Source of type %s is unknown" % src_type)

        else:

            # Store the type string
            self._src_type = str(src_type)

    def has_free_parameters(self):

        raise NotImplementedError("You need to override this")

    @property
    def free_parameters(self):
        """
        Returns a dictionary of free parameters for this source

        :return:
        """

        raise NotImplementedError("You need to override this")

    @property
    def components(self):
        """
        Return the dictionary of components

        :return: dictionary of components
        """

        return self._components

    @property
    def source_type(self):
        """
        Return the type of the source ('point source' or 'extended source')

        :return: type of the source
        """

        return self._src_type
